@@ -74,6 +74,11 @@ package main
 //#define PF_QIPCRTR AF_QIPCRTR
 //#endif				// AF_QIPCRTR
 //
+//#ifndef AF_XDP
+//#define AF_XDP 44
+//#define PF_XDP AF_XDP
+//#endif				// AF_XDP
+//
 // // https://github.com/sctplab/usrsctp/blob/master/usrsctplib/usrsctp.h
 //#ifndef AF_CONN
 //#define AF_CONN 123
@@ -128,6 +133,10 @@ package main
 //#define SCMP_ARCH_S390X ARCH_BAD
 //#endif
 //
+//#ifndef SCMP_ARCH_RISCV64
+//#define SCMP_ARCH_RISCV64 ARCH_BAD
+//#endif
+//
 //#ifndef SECCOMP_RET_LOG
 //#define SECCOMP_RET_LOG 0x7ffc0000U
 //#endif
@@ -185,8 +194,7 @@ import (
 	"strings"
 	"syscall"
 
-	// FIXME: we want github.com/seccomp/libseccomp-golang but that will not work with trusty because libseccomp-golang checks for the seccomp version and errors if it find one < 2.2.0
-	"github.com/mvo5/libseccomp-golang"
+	"github.com/seccomp/libseccomp-golang"
 
 	"github.com/snapcore/snapd/arch"
 	"github.com/snapcore/snapd/osutil"
@@ -283,6 +291,8 @@ var seccompResolver = map[string]uint64{
 	"PF_CONN":    C.PF_CONN,
 	"AF_QIPCRTR": C.AF_QIPCRTR, // 42
 	"PF_QIPCRTR": C.PF_QIPCRTR,
+	"AF_XDP":     C.AF_XDP, // 44
+	"PF_XDP":     C.PF_XDP,
 
 	// man 2 socket - type
 	"SOCK_STREAM":    syscall.SOCK_STREAM,
@@ -452,7 +462,7 @@ func DpkgArchToScmpArch(dpkgArch string) seccomp.ScmpArch {
 	case "s390x":
 		return seccomp.ArchS390X
 	}
-	panic(fmt.Sprintf("cannot map dpkg arch %q to a seccomp arch", dpkgArch))
+	return extraDpkgArchToScmpArch(dpkgArch)
 }
 
 // important for unit testing
@@ -476,25 +486,26 @@ func (sc *SeccompData) SetArgs(args [6]uint64) {
 // the arg is known to be 32 bit (uid_t/gid_t) and the kernel accepts one
 // or both of uint32(-1) and uint64(-1) and does its own masking).
 var syscallsWithNegArgsMaskHi32 = map[string]bool{
-	"chown":       true,
-	"chown32":     true,
-	"fchown":      true,
-	"fchown32":    true,
-	"fchownat":    true,
-	"lchown":      true,
-	"lchown32":    true,
-	"setgid":      true,
-	"setgid32":    true,
-	"setregid":    true,
-	"setregid32":  true,
-	"setresgid":   true,
-	"setresgid32": true,
-	"setreuid":    true,
-	"setreuid32":  true,
-	"setresuid":   true,
-	"setresuid32": true,
-	"setuid":      true,
-	"setuid32":    true,
+	"chown":           true,
+	"chown32":         true,
+	"fchown":          true,
+	"fchown32":        true,
+	"fchownat":        true,
+	"lchown":          true,
+	"lchown32":        true,
+	"setgid":          true,
+	"setgid32":        true,
+	"setregid":        true,
+	"setregid32":      true,
+	"setresgid":       true,
+	"setresgid32":     true,
+	"setreuid":        true,
+	"setreuid32":      true,
+	"setresuid":       true,
+	"setresuid32":     true,
+	"setuid":          true,
+	"setuid32":        true,
+	"copy_file_range": true,
 }
 
 // The kernel uses uint32 for all syscall arguments, but seccomp takes a
@@ -726,7 +737,7 @@ func complainAction() seccomp.ScmpAction {
 	}
 
 	// Because ActLog is functionally ActAllow with logging, if we don't
-	// support ActLog, fallback to ActLog.
+	// support ActLog, fallback to ActAllow.
 	return seccomp.ActAllow
 }
 
