@@ -103,6 +103,7 @@ type fakeQuotaGroupPostHandlerOpts struct {
 	groupName     string
 	parentName    string
 	snaps         []string
+	services      []string
 	maxMemory     int64
 	maxThreads    int
 	cpuCount      int
@@ -131,6 +132,7 @@ type quotasEnsureBody struct {
 	GroupName   string                      `json:"group-name,omitempty"`
 	ParentName  string                      `json:"parent,omitempty"`
 	Snaps       []string                    `json:"snaps,omitempty"`
+	Services    []string                    `json:"services,omitempty"`
 	Constraints quotasEnsureBodyConstraints `json:"constraints,omitempty"`
 }
 
@@ -152,6 +154,7 @@ func (s *quotaSuite) makeFakeQuotaPostHandler(c *check.C, opts fakeQuotaGroupPos
 				GroupName:   opts.groupName,
 				ParentName:  opts.parentName,
 				Snaps:       opts.snaps,
+				Services:    opts.services,
 				Constraints: quotasEnsureBodyConstraints{},
 			}
 			if opts.maxMemory != 0 {
@@ -223,6 +226,7 @@ func (s *quotaSuite) TestParseQuotas(c *check.C) {
 		{journalRateLimit: "10/15s", quotas: `{"journal":{"rate-count":10,"rate-period":15000000000}}`},
 		{journalRateLimit: "1500/15ms", quotas: `{"journal":{"rate-count":1500,"rate-period":15000000}}`},
 		{journalRateLimit: "1/15us", quotas: `{"journal":{"rate-count":1,"rate-period":15000}}`},
+		{journalRateLimit: "0/0s", quotas: `{"journal":{"rate-count":0,"rate-period":0}}`},
 
 		// Error cases
 		{cpuMax: "ASD", err: `cannot parse cpu quota string "ASD"`},
@@ -285,13 +289,13 @@ func (s *quotaSuite) TestSetQuotaCpuHappy(c *check.C) {
 		cpuCount:      2,
 		cpuPercentage: 50,
 	}
-	const getJsonTemplate = `{
+	// this data is not tested against, but it should still be valid
+	const getJson = `{
 		"type": "sync",
 		"status-code": 200,
 		"result": {
 			"group-name":"foo",
-			"constraints": { "memory": %d },
-			"current": { "memory": 500 }
+			"constraints": { "cpu":{"count":2, "percentage":50} },
 		}
 	}`
 	routes := map[string]http.HandlerFunc{
@@ -299,7 +303,7 @@ func (s *quotaSuite) TestSetQuotaCpuHappy(c *check.C) {
 			c,
 			fakeHandlerOpts,
 		),
-		"/v2/quotas/foo": s.makeFakeGetQuotaGroupHandler(c, fmt.Sprintf(getJsonTemplate, 1000)),
+		"/v2/quotas/foo": s.makeFakeGetQuotaGroupHandler(c, getJson),
 		"/v2/changes/42": makeChangesHandler(c),
 	}
 	s.RedirectClientToTestServer(dispatchFakeHandlers(c, routes))
@@ -311,6 +315,43 @@ func (s *quotaSuite) TestSetQuotaCpuHappy(c *check.C) {
 	c.Check(s.quotaPostHandlerCalls, check.Equals, 1)
 }
 
+func (s *quotaSuite) TestSetQuotaSnapServices(c *check.C) {
+	const postJSON = `{"type": "async", "status-code": 202,"change":"42", "result": []}`
+	fakeHandlerOpts := fakeQuotaGroupPostHandlerOpts{
+		action:        "ensure",
+		body:          postJSON,
+		groupName:     "foo",
+		snaps:         []string{"my-snap"},
+		services:      []string{"snap.svc1", "snap.svc2"},
+		cpuCount:      2,
+		cpuPercentage: 50,
+	}
+	// this data is not tested against, but it should still be valid
+	const getJson = `{
+		"type": "sync",
+		"status-code": 200,
+		"result": {
+			"group-name":"foo",
+			"constraints": { "cpu":{"count":2, "percentage":50} },
+		}
+	}`
+	routes := map[string]http.HandlerFunc{
+		"/v2/quotas": s.makeFakeQuotaPostHandler(
+			c,
+			fakeHandlerOpts,
+		),
+		"/v2/quotas/foo": s.makeFakeGetQuotaGroupHandler(c, getJson),
+		"/v2/changes/42": makeChangesHandler(c),
+	}
+	s.RedirectClientToTestServer(dispatchFakeHandlers(c, routes))
+
+	// ensure we correctly parse the snap.service format and send it to the daemon.
+	_, err := main.Parser(main.Client()).ParseArgs([]string{"set-quota", "--cpu=2x50%", "foo", "my-snap", "snap.svc1", "snap.svc2"})
+	c.Check(err, check.IsNil)
+	c.Check(s.quotaGetGroupHandlerCalls, check.Equals, 1)
+	c.Check(s.quotaPostHandlerCalls, check.Equals, 1)
+}
+
 func (s *quotaSuite) TestGetQuotaGroup(c *check.C) {
 	restore := main.MockIsStdinTTY(true)
 	defer restore()
@@ -323,6 +364,7 @@ func (s *quotaSuite) TestGetQuotaGroup(c *check.C) {
 			"parent":"bar",
 			"subgroups":["subgrp1"],
 			"snaps":["snap-a","snap-b"],
+			"services":["snap-a.svc1", "snap-b.svc2"],
 			"constraints": { "memory": 1000 },
 			"current": { "memory": 900 }
 		}
@@ -346,6 +388,9 @@ subgroups:
 snaps:
   - snap-a
   - snap-b
+services:
+  - snap-a.svc1
+  - snap-b.svc2
 `[1:])
 	c.Check(s.quotaGetGroupHandlerCalls, check.Equals, 1)
 	c.Check(s.quotaPostHandlerCalls, check.Equals, 0)
@@ -660,11 +705,12 @@ func (s *quotaSuite) TestGetAllQuotaGroups(c *check.C) {
 			{"group-name":"fff","parent":"aaa","constraints":{"memory":1000},"current":{"memory":0}},
 			{"group-name":"xxx","constraints":{"memory":9900},"current":{"memory":10000}},
 			{"group-name":"cp0","constraints":{"memory":9900, "cpu":{"percentage":90}},"current":{"memory":10000}},
-			{"group-name":"cp1","subgroups":["cps0","js0"],"constraints":{"cpu":{"count":2, "percentage":90}}},
+			{"group-name":"cp1","subgroups":["cps0","js0","js1"],"constraints":{"cpu":{"count":2, "percentage":90}}},
 			{"group-name":"cps0","parent":"cp1","constraints":{"cpu":{"percentage":40}}},
 			{"group-name":"cp2","subgroups":["cps1"],"constraints":{"cpu":{"count":2,"percentage":100},"cpu-set":{"cpus":[0,1]}}},
 			{"group-name":"cps1","parent":"cp2","constraints":{"memory":9900,"cpu":{"percentage":50},"cpu-set":{"cpus":[1]}},"current":{"memory":10000}},
-			{"group-name":"js0","parent":"cp1","constraints":{"journal":{"size":1048576,"rate-count":50,"rate-period":60000000000}}}
+			{"group-name":"js0","parent":"cp1","constraints":{"journal":{"size":1048576,"rate-count":50,"rate-period":60000000000}}},
+			{"group-name":"js1","parent":"cp1","constraints":{"journal":{"rate-count":0,"rate-period":0}}}
 			]}`))
 
 	rest, err := main.Parser(main.Client()).ParseArgs([]string{"quotas"})
@@ -677,6 +723,7 @@ cp0              memory=9.9kB,cpu=90%                      memory=10.0kB
 cp1              cpu=2x90%                                 
 cps0     cp1     cpu=40%                                   
 js0      cp1     journal-size=1.05MB,journal-rate=50/1m0s  
+js1      cp1     journal-rate=0/0s                         
 cp2              cpu=2x100%,cpu-set=0,1                    
 cps1     cp2     memory=9.9kB,cpu=50%,cpu-set=1            memory=10.0kB
 ggg              memory=1000B,threads=100                  memory=3000B
