@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2016-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,6 +20,7 @@
 package main_test
 
 import (
+	"encoding/json"
 	"errors"
 	"fmt"
 	"io/ioutil"
@@ -32,12 +33,16 @@ import (
 
 	"gopkg.in/check.v1"
 
+	"github.com/godbus/dbus"
 	snaprun "github.com/snapcore/snapd/cmd/snap"
 	"github.com/snapcore/snapd/cmd/snaplock/runinhibit"
+	"github.com/snapcore/snapd/dbusutil"
+	"github.com/snapcore/snapd/dbusutil/dbustest"
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/features"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/osutil/strace"
 	"github.com/snapcore/snapd/progress"
 	"github.com/snapcore/snapd/progress/progresstest"
 	"github.com/snapcore/snapd/sandbox/cgroup"
@@ -45,6 +50,7 @@ import (
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/snap/snaptest"
 	"github.com/snapcore/snapd/testutil"
+	usersessionclient "github.com/snapcore/snapd/usersession/client"
 	"github.com/snapcore/snapd/x11"
 )
 
@@ -232,7 +238,7 @@ func (s *RunSuite) TestSnapRunAppRunsChecksInhibitionLock(c *check.C) {
 	c.Assert(ioutil.WriteFile(features.RefreshAppAwareness.ControlFile(), []byte(nil), 0644), check.IsNil)
 
 	var called int
-	restore := snaprun.MockWaitInhibitUnlock(func(snapName string, waitFor runinhibit.Hint, errCh <-chan error) (bool, error) {
+	restore := snaprun.MockWaitInhibitUnlock(func(snapName string, waitFor runinhibit.Hint) (bool, error) {
 		called++
 		return false, nil
 	})
@@ -271,7 +277,7 @@ func (s *RunSuite) TestSnapRunHookNoRuninhibit(c *check.C) {
 	defer restorer()
 
 	var called bool
-	restore := snaprun.MockWaitInhibitUnlock(func(snapName string, waitFor runinhibit.Hint, errCh <-chan error) (bool, error) {
+	restore := snaprun.MockWaitInhibitUnlock(func(snapName string, waitFor runinhibit.Hint) (bool, error) {
 		called = true
 		c.Errorf("WaitInhibitUnlock should not have been called")
 		return false, nil
@@ -315,7 +321,7 @@ func (s *RunSuite) TestSnapRunAppRuninhibitSkipsServices(c *check.C) {
 	c.Assert(ioutil.WriteFile(features.RefreshAppAwareness.ControlFile(), []byte(nil), 0644), check.IsNil)
 
 	var called bool
-	restore := snaprun.MockWaitInhibitUnlock(func(snapName string, waitFor runinhibit.Hint, errCh <-chan error) (bool, error) {
+	restore := snaprun.MockWaitInhibitUnlock(func(snapName string, waitFor runinhibit.Hint) (bool, error) {
 		called = true
 		c.Errorf("WaitInhibitUnlock should not have been called")
 		return false, nil
@@ -478,14 +484,37 @@ func (s *RunSuite) TestSnapRunAppWithCommandIntegration(c *check.C) {
 }
 
 func (s *RunSuite) TestSnapRunCreateDataDirs(c *check.C) {
+	for _, t := range []struct {
+		snapDir string
+		opts    *dirs.SnapDirOptions
+	}{
+		{snapDir: dirs.UserHomeSnapDir},
+		{snapDir: dirs.UserHomeSnapDir, opts: &dirs.SnapDirOptions{}},
+		{snapDir: dirs.HiddenSnapDataHomeDir, opts: &dirs.SnapDirOptions{HiddenSnapDataDir: true}},
+	} {
+		s.testSnapRunCreateDataDirs(c, t.snapDir, t.opts)
+		c.Assert(os.RemoveAll(s.fakeHome), check.IsNil)
+		s.fakeHome = c.MkDir()
+	}
+}
+
+func (s *RunSuite) testSnapRunCreateDataDirs(c *check.C, snapDir string, opts *dirs.SnapDirOptions) {
 	info, err := snap.InfoFromSnapYaml(mockYaml)
 	c.Assert(err, check.IsNil)
 	info.SideInfo.Revision = snap.R(42)
 
-	err = snaprun.CreateUserDataDirs(info)
+	err = snaprun.CreateUserDataDirs(info, opts)
 	c.Assert(err, check.IsNil)
-	c.Check(osutil.FileExists(filepath.Join(s.fakeHome, "/snap/snapname/42")), check.Equals, true)
-	c.Check(osutil.FileExists(filepath.Join(s.fakeHome, "/snap/snapname/common")), check.Equals, true)
+	c.Check(osutil.FileExists(filepath.Join(s.fakeHome, snapDir, "snapname/42")), check.Equals, true)
+	c.Check(osutil.FileExists(filepath.Join(s.fakeHome, snapDir, "snapname/common")), check.Equals, true)
+
+	// check we don't create the alternative dir
+	nonExistentDir := dirs.HiddenSnapDataHomeDir
+	if snapDir == dirs.HiddenSnapDataHomeDir {
+		nonExistentDir = dirs.UserHomeSnapDir
+	}
+
+	c.Check(osutil.FileExists(filepath.Join(s.fakeHome, nonExistentDir)), check.Equals, false)
 }
 
 func (s *RunSuite) TestParallelInstanceSnapRunCreateDataDirs(c *check.C) {
@@ -494,7 +523,7 @@ func (s *RunSuite) TestParallelInstanceSnapRunCreateDataDirs(c *check.C) {
 	info.SideInfo.Revision = snap.R(42)
 	info.InstanceKey = "foo"
 
-	err = snaprun.CreateUserDataDirs(info)
+	err = snaprun.CreateUserDataDirs(info, nil)
 	c.Assert(err, check.IsNil)
 	c.Check(osutil.FileExists(filepath.Join(s.fakeHome, "/snap/snapname_foo/42")), check.Equals, true)
 	c.Check(osutil.FileExists(filepath.Join(s.fakeHome, "/snap/snapname_foo/common")), check.Equals, true)
@@ -993,7 +1022,7 @@ echo "stdout output 2"
 			filepath.Join(straceCmd.BinDir(), "strace"),
 			"-u", user.Username,
 			"-f",
-			"-e", "!select,pselect6,_newselect,clock_gettime,sigaltstack,gettid,gettimeofday,nanosleep",
+			"-e", strace.ExcludedSyscalls,
 			filepath.Join(dirs.DistroLibExecDir, "snap-confine"),
 			"snap.snapname.app",
 			filepath.Join(dirs.CoreLibExecDir, "snap-exec"),
@@ -1016,7 +1045,7 @@ echo "stdout output 2"
 			filepath.Join(straceCmd.BinDir(), "strace"),
 			"-u", user.Username,
 			"-f",
-			"-e", "!select,pselect6,_newselect,clock_gettime,sigaltstack,gettid,gettimeofday,nanosleep",
+			"-e", strace.ExcludedSyscalls,
 			filepath.Join(dirs.DistroLibExecDir, "snap-confine"),
 			"snap.snapname.app",
 			filepath.Join(dirs.CoreLibExecDir, "snap-exec"),
@@ -1063,7 +1092,7 @@ func (s *RunSuite) TestSnapRunAppWithStraceOptions(c *check.C) {
 			filepath.Join(straceCmd.BinDir(), "strace"),
 			"-u", user.Username,
 			"-f",
-			"-e", "!select,pselect6,_newselect,clock_gettime,sigaltstack,gettid,gettimeofday,nanosleep",
+			"-e", strace.ExcludedSyscalls,
 			"-tt",
 			"-o",
 			"file with spaces",
@@ -1580,6 +1609,68 @@ func (s *RunSuite) TestSnapRunTrackingServices(c *check.C) {
 	c.Assert(confirmed, check.Equals, true)
 }
 
+func (s *RunSuite) TestSnapRunTrackingServicesWhenRunByUser(c *check.C) {
+	restore := mockSnapConfine(filepath.Join(dirs.SnapMountDir, "core", "111", dirs.CoreLibExecDir))
+	defer restore()
+
+	// mock installed snap
+	snaptest.MockSnapCurrent(c, string(mockYaml), &snap.SideInfo{
+		Revision: snap.R("x2"),
+	})
+
+	// pretend to be running from core
+	restore = snaprun.MockOsReadlink(func(string) (string, error) {
+		return filepath.Join(dirs.SnapMountDir, "core/111/usr/bin/snap"), nil
+	})
+	defer restore()
+
+	var createTransientScopeOpts *cgroup.TrackingOptions
+	var createTransientScopeCalls int
+	restore = snaprun.MockCreateTransientScopeForTracking(func(securityTag string, opts *cgroup.TrackingOptions) error {
+		createTransientScopeCalls++
+		createTransientScopeOpts = opts
+		return nil
+	})
+	defer restore()
+
+	confirmCalls := 0
+	restore = snaprun.MockConfirmSystemdServiceTracking(func(securityTag string) error {
+		confirmCalls++
+		c.Assert(securityTag, check.Equals, "snap.snapname.svc")
+		return cgroup.ErrCannotTrackProcess
+	})
+	defer restore()
+
+	// redirect exec
+	execArg0 := ""
+	execArgs := []string{}
+	execEnv := []string{}
+	restore = snaprun.MockSyscallExec(func(arg0 string, args []string, envv []string) error {
+		execArg0 = arg0
+		execArgs = args
+		execEnv = envv
+		return nil
+	})
+	defer restore()
+
+	// invoked as: snap run -- snapname.svc --arg1 arg2
+	rest, err := snaprun.Parser(snaprun.Client()).ParseArgs([]string{"run", "--", "snapname.svc", "--arg1", "arg2"})
+	c.Assert(err, check.IsNil)
+	c.Assert(rest, check.DeepEquals, []string{"snapname.svc", "--arg1", "arg2"})
+	c.Check(execArg0, check.Equals, filepath.Join(dirs.SnapMountDir, "/core/111", dirs.CoreLibExecDir, "snap-confine"))
+	c.Check(execArgs, check.DeepEquals, []string{
+		filepath.Join(dirs.SnapMountDir, "/core/111", dirs.CoreLibExecDir, "snap-confine"),
+		"snap.snapname.svc",
+		filepath.Join(dirs.CoreLibExecDir, "snap-exec"),
+		"snapname.svc", "--arg1", "arg2"})
+	c.Check(execEnv, testutil.Contains, "SNAP_REVISION=x2")
+	c.Assert(confirmCalls, check.Equals, 1)
+	c.Assert(createTransientScopeCalls, check.Equals, 1)
+	c.Assert(createTransientScopeOpts, check.DeepEquals, &cgroup.TrackingOptions{
+		AllowSessionBus: true,
+	})
+}
+
 func (s *RunSuite) TestSnapRunTrackingFailure(c *check.C) {
 	restore := mockSnapConfine(filepath.Join(dirs.SnapMountDir, "core", "111", dirs.CoreLibExecDir))
 	defer restore()
@@ -1664,7 +1755,7 @@ func (s *RunSuite) TestSnapRunHookKernelImplicitBase(c *check.C) {
 			case 0:
 				c.Check(r.Method, check.Equals, "GET")
 				c.Check(r.URL.RawQuery, check.Equals, "")
-				fmt.Fprintln(w, happyUC20ModelAssertionResponse)
+				fmt.Fprint(w, happyUC20ModelAssertionResponse)
 			default:
 				c.Fatalf("expected to get 1 request for /v2/model, now on %d", nModel+1)
 			}
@@ -1728,7 +1819,7 @@ func (s *RunSuite) TestWaitInhibitUnlock(c *check.C) {
 	})
 	defer restore()
 
-	notInhibited, err := snaprun.WaitInhibitUnlock("some-snap", runinhibit.HintNotInhibited, nil)
+	notInhibited, err := snaprun.WaitInhibitUnlock("some-snap", runinhibit.HintNotInhibited)
 	c.Assert(err, check.IsNil)
 	c.Check(notInhibited, check.Equals, true)
 	c.Check(called, check.Equals, 5)
@@ -1745,31 +1836,12 @@ func (s *RunSuite) TestWaitInhibitUnlockWaitsForSpecificHint(c *check.C) {
 	})
 	defer restore()
 
-	notInhibited, err := snaprun.WaitInhibitUnlock("some-snap", runinhibit.HintInhibitedForRefresh, nil)
+	notInhibited, err := snaprun.WaitInhibitUnlock("some-snap", runinhibit.HintInhibitedForRefresh)
 	c.Assert(err, check.IsNil)
 	c.Check(notInhibited, check.Equals, false)
 	c.Check(called, check.Equals, 5)
 }
 
-func (s *RunSuite) TestWaitInhibitUnlockWithErrorChannel(c *check.C) {
-	errCh := make(chan error, 1)
-	var called int
-	restore := snaprun.MockIsLocked(func(snapName string) (runinhibit.Hint, error) {
-		called++
-		if called == 1 {
-			errCh <- fmt.Errorf("boom")
-		}
-		return runinhibit.HintInhibitedForRefresh, nil
-	})
-	defer restore()
-
-	notInhibited, err := snaprun.WaitInhibitUnlock("some-snap", runinhibit.HintNotInhibited, errCh)
-	c.Assert(err, check.IsNil)
-	c.Check(notInhibited, check.Equals, false)
-	c.Check(called, check.Equals, 1)
-	c.Check(s.Stderr(), check.Equals, `boom`)
-}
-
 func (s *RunSuite) TestWaitWhileInhibitedNoop(c *check.C) {
 	var called int
 	restore := snaprun.MockIsLocked(func(snapName string) (runinhibit.Hint, error) {
@@ -1819,3 +1891,297 @@ func (s *RunSuite) TestWaitWhileInhibitedTextFlow(c *check.C) {
 	c.Check(meter.Finishes, check.Equals, 1)
 	c.Check(meter.Labels, check.DeepEquals, []string{"please wait..."})
 }
+
+func (s *RunSuite) TestWaitWhileInhibitedGraphicalSessionFlow(c *check.C) {
+	_, r := logger.MockLogger()
+	defer r()
+
+	restoreIsGraphicalSession := snaprun.MockIsGraphicalSession(true)
+	defer restoreIsGraphicalSession()
+
+	restoreTryNotifyRefresh := snaprun.MockTryNotifyRefreshViaSnapDesktopIntegrationFlow(func(snapName string) (bool, error) {
+		c.Check(snapName, check.Equals, "some-snap")
+		return false, nil
+	})
+	defer restoreTryNotifyRefresh()
+
+	restorePendingRefreshNotification := snaprun.MockPendingRefreshNotification(func(refreshInfo *usersessionclient.PendingSnapRefreshInfo) error {
+		c.Check(refreshInfo, check.DeepEquals, &usersessionclient.PendingSnapRefreshInfo{
+			InstanceName:  "some-snap",
+			TimeRemaining: 0,
+		})
+		return nil
+	})
+	defer restorePendingRefreshNotification()
+
+	restoreFinishRefreshNotification := snaprun.MockFinishRefreshNotification(func(refreshInfo *usersessionclient.FinishedSnapRefreshInfo) error {
+		c.Check(refreshInfo, check.DeepEquals, &usersessionclient.FinishedSnapRefreshInfo{
+			InstanceName: "some-snap",
+		})
+		return nil
+	})
+	defer restoreFinishRefreshNotification()
+
+	var called int
+	restore := snaprun.MockIsLocked(func(snapName string) (runinhibit.Hint, error) {
+		c.Check(snapName, check.Equals, "some-snap")
+		called++
+		if called < 2 {
+			return runinhibit.HintInhibitedForRefresh, nil
+		}
+		return runinhibit.HintNotInhibited, nil
+	})
+	defer restore()
+
+	c.Assert(runinhibit.LockWithHint("some-snap", runinhibit.HintInhibitedForRefresh), check.IsNil)
+	c.Assert(snaprun.WaitWhileInhibited("some-snap"), check.IsNil)
+	c.Check(called, check.Equals, 2)
+	c.Check(s.Stdout(), check.Equals, "")
+}
+
+func (s *RunSuite) TestWaitWhileInhibitedGraphicalSessionFlowError(c *check.C) {
+	_, r := logger.MockLogger()
+	defer r()
+
+	restoreIsGraphicalSession := snaprun.MockIsGraphicalSession(true)
+	defer restoreIsGraphicalSession()
+
+	restoreTryNotifyRefresh := snaprun.MockTryNotifyRefreshViaSnapDesktopIntegrationFlow(func(snapName string) (bool, error) {
+		c.Check(snapName, check.Equals, "some-snap")
+		return false, nil
+	})
+	defer restoreTryNotifyRefresh()
+
+	restorePendingRefreshNotification := snaprun.MockPendingRefreshNotification(func(refreshInfo *usersessionclient.PendingSnapRefreshInfo) error {
+		c.Check(refreshInfo, check.DeepEquals, &usersessionclient.PendingSnapRefreshInfo{
+			InstanceName:  "some-snap",
+			TimeRemaining: 0,
+		})
+		return fmt.Errorf("boom")
+	})
+	defer restorePendingRefreshNotification()
+
+	c.Assert(runinhibit.LockWithHint("some-snap", runinhibit.HintInhibitedForRefresh), check.IsNil)
+	restore := snaprun.MockIsLocked(func(snapName string) (runinhibit.Hint, error) {
+		c.Check(snapName, check.Equals, "some-snap")
+		return runinhibit.HintInhibitedForRefresh, nil
+	})
+	defer restore()
+
+	c.Assert(snaprun.WaitWhileInhibited("some-snap"), check.ErrorMatches, "boom")
+}
+
+func (s *RunSuite) TestWaitWhileInhibitedGraphicalSessionFlowErrorOnFinish(c *check.C) {
+	_, r := logger.MockLogger()
+	defer r()
+
+	restoreIsGraphicalSession := snaprun.MockIsGraphicalSession(true)
+	defer restoreIsGraphicalSession()
+
+	restoreTryNotifyRefresh := snaprun.MockTryNotifyRefreshViaSnapDesktopIntegrationFlow(func(snapName string) (bool, error) {
+		c.Check(snapName, check.Equals, "some-snap")
+		return false, nil
+	})
+	defer restoreTryNotifyRefresh()
+
+	restorePendingRefreshNotification := snaprun.MockPendingRefreshNotification(func(refreshInfo *usersessionclient.PendingSnapRefreshInfo) error {
+		c.Check(refreshInfo, check.DeepEquals, &usersessionclient.PendingSnapRefreshInfo{
+			InstanceName:  "some-snap",
+			TimeRemaining: 0,
+		})
+		return nil
+	})
+	defer restorePendingRefreshNotification()
+
+	restoreFinishRefreshNotification := snaprun.MockFinishRefreshNotification(func(refreshInfo *usersessionclient.FinishedSnapRefreshInfo) error {
+		c.Check(refreshInfo, check.DeepEquals, &usersessionclient.FinishedSnapRefreshInfo{
+			InstanceName: "some-snap",
+		})
+		return fmt.Errorf("boom")
+	})
+	defer restoreFinishRefreshNotification()
+
+	c.Assert(runinhibit.LockWithHint("some-snap", runinhibit.HintInhibitedForRefresh), check.IsNil)
+	n := 0
+	restore := snaprun.MockIsLocked(func(snapName string) (runinhibit.Hint, error) {
+		c.Check(snapName, check.Equals, "some-snap")
+		n++
+		if n == 1 {
+			return runinhibit.HintInhibitedForRefresh, nil
+		}
+		return runinhibit.HintNotInhibited, nil
+	})
+	defer restore()
+
+	c.Assert(snaprun.WaitWhileInhibited("some-snap"), check.ErrorMatches, "boom")
+}
+
+func (s *RunSuite) TestCreateSnapDirPermissions(c *check.C) {
+	usr, err := user.Current()
+	c.Assert(err, check.IsNil)
+
+	usr.HomeDir = s.fakeHome
+	snaprun.MockUserCurrent(func() (*user.User, error) {
+		return usr, nil
+	})
+
+	info := &snap.Info{SuggestedName: "some-snap"}
+	c.Assert(snaprun.CreateUserDataDirs(info, nil), check.IsNil)
+
+	fi, err := os.Stat(filepath.Join(s.fakeHome, dirs.UserHomeSnapDir))
+	c.Assert(err, check.IsNil)
+	c.Assert(fi.Mode()&os.ModePerm, check.Equals, os.FileMode(0700))
+}
+
+func (s *RunSuite) TestGetSnapDirOptions(c *check.C) {
+	root := c.MkDir()
+	dirs.SnapSeqDir = root
+	dirs.FeaturesDir = root
+
+	// write sequence file
+	seqFile := filepath.Join(dirs.SnapSeqDir, "somesnap.json")
+	str := struct {
+		MigratedHidden        bool `json:"migrated-hidden"`
+		MigratedToExposedHome bool `json:"migrated-exposed-home"`
+	}{
+		MigratedHidden:        true,
+		MigratedToExposedHome: true,
+	}
+	data, err := json.Marshal(&str)
+	c.Assert(err, check.IsNil)
+	c.Assert(ioutil.WriteFile(seqFile, data, 0660), check.IsNil)
+
+	// write control file for hidden dir feature
+	c.Assert(ioutil.WriteFile(features.HiddenSnapDataHomeDir.ControlFile(), []byte{}, 0660), check.IsNil)
+
+	opts, err := snaprun.GetSnapDirOptions("somesnap")
+	c.Assert(err, check.IsNil)
+	c.Assert(opts, check.DeepEquals, &dirs.SnapDirOptions{HiddenSnapDataDir: true, MigratedToExposedHome: true})
+}
+
+func (s *RunSuite) TestRunDebugLog(c *check.C) {
+	oldDebug, isSet := os.LookupEnv("SNAPD_DEBUG")
+	if isSet {
+		defer os.Setenv("SNAPD_DEBUG", oldDebug)
+	} else {
+		defer os.Unsetenv("SNAPD_DEBUG")
+	}
+
+	logBuf, r := logger.MockLogger()
+	defer r()
+
+	restore := mockSnapConfine(dirs.DistroLibExecDir)
+	defer restore()
+	execArg0 := ""
+	execArgs := []string{}
+	execEnv := []string{}
+	restore = snaprun.MockSyscallExec(func(arg0 string, args []string, envv []string) error {
+		execArg0 = arg0
+		execArgs = args
+		execEnv = envv
+		return nil
+	})
+	defer restore()
+
+	snaptest.MockSnapCurrent(c, string(mockYaml), &snap.SideInfo{
+		Revision: snap.R("12"),
+	})
+
+	// this will modify the current process environment
+	_, err := snaprun.Parser(snaprun.Client()).ParseArgs([]string{"run", "--debug-log", "snapname.app"})
+	c.Assert(err, check.IsNil)
+	c.Check(execArg0, check.Equals, filepath.Join(dirs.DistroLibExecDir, "snap-confine"))
+	c.Check(execArgs, check.DeepEquals, []string{
+		filepath.Join(dirs.DistroLibExecDir, "snap-confine"),
+		"snap.snapname.app",
+		filepath.Join(dirs.CoreLibExecDir, "snap-exec"),
+		"snapname.app"})
+	c.Check(execEnv, testutil.Contains, "SNAP_REVISION=12")
+	c.Check(execEnv, testutil.Contains, "SNAPD_DEBUG=1")
+	// also set in env
+	c.Check(os.Getenv("SNAPD_DEBUG"), check.Equals, "1")
+	// and we've let the user know that logging was enabled
+	c.Check(logBuf.String(), testutil.Contains, "DEBUG: enabled debug logging of early snap startup")
+}
+
+func (s *RunSuite) TestDesktopIntegrationNoDBus(c *check.C) {
+	_, r := logger.MockLogger()
+	defer r()
+
+	noDBus := func() (*dbus.Conn, error) { return nil, fmt.Errorf("dbus not available") }
+	restore := dbusutil.MockConnections(noDBus, noDBus)
+	defer restore()
+
+	sent, err := snaprun.TryNotifyRefreshViaSnapDesktopIntegrationFlow("Test")
+	c.Assert(sent, check.Equals, false)
+	c.Assert(err, check.IsNil)
+}
+
+func makeDBusMethodNotAvailableMessage(c *check.C, msg *dbus.Message) *dbus.Message {
+	return &dbus.Message{
+		Type: dbus.TypeError,
+		Headers: map[dbus.HeaderField]dbus.Variant{
+			dbus.FieldReplySerial: dbus.MakeVariant(msg.Serial()),
+			dbus.FieldSender:      dbus.MakeVariant(":1"), // This does not matter.
+			// dbus.FieldDestination is provided automatically by DBus test helper.
+			dbus.FieldErrorName: dbus.MakeVariant("org.freedesktop.DBus.Error.UnknownMethod"),
+		},
+	}
+}
+
+func (s *RunSuite) TestDesktopIntegrationDBusAvailableNoMethod(c *check.C) {
+	_, r := logger.MockLogger()
+	defer r()
+
+	conn, _, err := dbustest.InjectableConnection(func(msg *dbus.Message, n int) ([]*dbus.Message, error) {
+		return []*dbus.Message{makeDBusMethodNotAvailableMessage(c, msg)}, nil
+	})
+	c.Assert(err, check.IsNil)
+
+	restore := dbusutil.MockOnlySessionBusAvailable(conn)
+	defer restore()
+
+	sent, err := snaprun.TryNotifyRefreshViaSnapDesktopIntegrationFlow("SnapTest")
+	c.Assert(sent, check.Equals, false)
+	c.Assert(err, check.IsNil)
+}
+
+func makeDBusMethodAvailableMessage(c *check.C, msg *dbus.Message) *dbus.Message {
+	c.Assert(msg.Type, check.Equals, dbus.TypeMethodCall)
+	c.Check(msg.Flags, check.Equals, dbus.Flags(0))
+
+	c.Check(msg.Headers, check.DeepEquals, map[dbus.HeaderField]dbus.Variant{
+		dbus.FieldDestination: dbus.MakeVariant("io.snapcraft.SnapDesktopIntegration"),
+		dbus.FieldPath:        dbus.MakeVariant(dbus.ObjectPath("/io/snapcraft/SnapDesktopIntegration")),
+		dbus.FieldInterface:   dbus.MakeVariant("io.snapcraft.SnapDesktopIntegration"),
+		dbus.FieldMember:      dbus.MakeVariant("ApplicationIsBeingRefreshed"),
+		dbus.FieldSignature:   dbus.MakeVariant(dbus.SignatureOf("", "", make(map[string]dbus.Variant))),
+	})
+	c.Check(msg.Body[0], check.Equals, "SnapTest")
+	param2 := fmt.Sprintf("%s", msg.Body[1])
+	c.Check(strings.HasSuffix(param2, "/var/lib/snapd/inhibit/SnapTest.lock"), check.Equals, true)
+	return &dbus.Message{
+		Type: dbus.TypeMethodReply,
+		Headers: map[dbus.HeaderField]dbus.Variant{
+			dbus.FieldReplySerial: dbus.MakeVariant(msg.Serial()),
+			dbus.FieldSender:      dbus.MakeVariant(":1"), // This does not matter.
+		},
+	}
+}
+
+func (s *RunSuite) TestDesktopIntegrationDBusAvailableMethodWorks(c *check.C) {
+	_, r := logger.MockLogger()
+	defer r()
+
+	conn, _, err := dbustest.InjectableConnection(func(msg *dbus.Message, n int) ([]*dbus.Message, error) {
+		return []*dbus.Message{makeDBusMethodAvailableMessage(c, msg)}, nil
+	})
+	c.Assert(err, check.IsNil)
+
+	restore := dbusutil.MockOnlySessionBusAvailable(conn)
+	defer restore()
+
+	sent, err := snaprun.TryNotifyRefreshViaSnapDesktopIntegrationFlow("SnapTest")
+	c.Assert(sent, check.Equals, true)
+	c.Assert(err, check.IsNil)
+}
