@@ -19,6 +19,18 @@
 
 package builtin
 
+import (
+	"path/filepath"
+
+	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/interfaces"
+	"github.com/snapcore/snapd/interfaces/apparmor"
+	"github.com/snapcore/snapd/interfaces/mount"
+	"github.com/snapcore/snapd/logger"
+	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/snap"
+)
+
 const systemObserveSummary = `allows observing all processes and drivers`
 
 const systemObserveBaseDeclarationSlots = `
@@ -53,6 +65,8 @@ ptrace (read),
 @{PROC}/cgroups r,
 @{PROC}/locks r,
 @{PROC}/modules r,
+@{PROC}/mdstat r,
+@{PROC}/schedstat r,
 @{PROC}/stat r,
 @{PROC}/vmstat r,
 @{PROC}/zoneinfo r,
@@ -96,6 +110,10 @@ ptrace (read),
 /var/lib/snapd/hostfs/etc/os-release rk,
 /var/lib/snapd/hostfs/usr/lib/os-release rk,
 
+# Allow discovering the Kernel build config
+@{PROC}/config.gz r,
+/boot/config* r,
+
 # Allow discovering system-wide CFS Bandwidth Control information
 # https://www.kernel.org/doc/html/latest/scheduler/sched-bwc.html
 /sys/fs/cgroup/cpu,cpuacct/cpu.cfs_period_us r,
@@ -126,7 +144,7 @@ dbus (send)
     bus={session,system}
     path=/org/freedesktop/DBus
     interface=org.freedesktop.DBus
-    member=ListNames
+    member={ListNames,ListActivatableNames}
     peer=(label=unconfined),
 
 # Allow clients to obtain the DBus machine ID on common buses. We do not
@@ -158,15 +176,53 @@ const systemObserveConnectedPlugSecComp = `
 #@deny ptrace
 `
 
+type systemObserveInterface struct {
+	commonInterface
+}
+
+func (iface *systemObserveInterface) AppArmorConnectedPlug(spec *apparmor.Specification, plug *interfaces.ConnectedPlug, slot *interfaces.ConnectedSlot) error {
+	spec.AddSnippet(systemObserveConnectedPlugAppArmor)
+	spec.SetSuppressPtraceTrace()
+	// Allow mounting boot partition to snap-update-ns
+	emit := spec.AddUpdateNSf
+	target := "/boot"
+	source := "/var/lib/snapd/hostfs" + target
+	emit("  # Read-only access to %s", target)
+	// When setting up a read-only bind mount, snap-update-ns first creates a
+	// plain read/write bind mount, and then remounts it to readonly.
+	emit("  mount options=(bind,rw) %s/ -> %s/,", source, target)
+	emit("  mount options=(bind,remount,ro) -> %s/,", target)
+	emit("  umount %s/,\n", target)
+	return nil
+}
+
+func (iface *systemObserveInterface) MountPermanentPlug(spec *mount.Specification, plug *snap.PlugInfo) error {
+	dir := filepath.Join(dirs.GlobalRootDir, "/boot")
+	if matches, _ := filepath.Glob(filepath.Join(dir, "config*")); len(matches) > 0 {
+		spec.AddMountEntry(osutil.MountEntry{
+			Name:    "/var/lib/snapd/hostfs" + dir,
+			Dir:     "/boot",
+			Options: []string{"bind", "ro"},
+		})
+	} else {
+		// TODO: if /boot/config does not exist, we should check whether the
+		// kernel is being delivered as a snap (this is the case in Ubuntu
+		// Core) and, if found, we should bind-mount the config file onto the
+		// expected location.
+		logger.Debugf("system-observe: /boot/config* not found, skipping mount of /boot/")
+	}
+	return nil
+}
+
 func init() {
-	registerIface(&commonInterface{
-		name:                  "system-observe",
-		summary:               systemObserveSummary,
-		implicitOnCore:        true,
-		implicitOnClassic:     true,
-		baseDeclarationSlots:  systemObserveBaseDeclarationSlots,
-		connectedPlugAppArmor: systemObserveConnectedPlugAppArmor,
-		connectedPlugSecComp:  systemObserveConnectedPlugSecComp,
-		suppressPtraceTrace:   true,
+	registerIface(&systemObserveInterface{
+		commonInterface: commonInterface{
+			name:                 "system-observe",
+			summary:              systemObserveSummary,
+			implicitOnCore:       true,
+			implicitOnClassic:    true,
+			baseDeclarationSlots: systemObserveBaseDeclarationSlots,
+			connectedPlugSecComp: systemObserveConnectedPlugSecComp,
+		},
 	})
 }
