@@ -23,6 +23,7 @@ import (
 	"errors"
 	"fmt"
 	"sort"
+	"strings"
 
 	tomb "gopkg.in/tomb.v2"
 
@@ -34,6 +35,7 @@ import (
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/progress"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/naming"
 	"github.com/snapcore/snapd/snap/quota"
 	"github.com/snapcore/snapd/snapdenv"
 	"github.com/snapcore/snapd/strutil"
@@ -56,6 +58,10 @@ type QuotaControlAction struct {
 	// the "update" or the "create" actions.
 	AddSnaps []string `json:"snaps,omitempty"`
 
+	// AddServices is the set of services to add to the quota group, valid for either
+	// the "update" or the "create" actions.
+	AddServices []string `json:"services,omitempty"`
+
 	// ResourceLimits is the set of resource limits to set on the quota group.
 	// Either the initial limit the group is created with for the "create"
 	// action, or if non-zero for the "update" the memory limit, then the new
@@ -178,8 +184,7 @@ func (m *ServiceManager) doQuotaControl(t *state.Task, _ *tomb.Tomb) error {
 	return nil
 }
 
-func addRefreshProfileTasks(st *state.State, queueTask func(task *state.Task), servicesAffected map[*snap.Info][]*snap.AppInfo) *state.TaskSet {
-	ts := state.NewTaskSet()
+func addRefreshProfileTasks(st *state.State, queueTask func(task *state.Task), servicesAffected map[*snap.Info][]*snap.AppInfo) {
 	for info := range servicesAffected {
 		setupProfilesTask := st.NewTask("setup-profiles", fmt.Sprintf(i18n.G("Update snap %q (%s) security profiles"), info.SnapName(), info.Revision))
 		setupProfilesTask.Set("snap-setup", &snapstate.SnapSetup{
@@ -190,7 +195,6 @@ func addRefreshProfileTasks(st *state.State, queueTask func(task *state.Task), s
 		})
 		queueTask(setupProfilesTask)
 	}
-	return ts
 }
 
 func addRestartServicesTasks(st *state.State, queueTask func(task *state.Task), grpName string, servicesAffected map[*snap.Info][]*snap.AppInfo) {
@@ -199,6 +203,7 @@ func addRestartServicesTasks(st *state.State, queueTask func(task *state.Task),
 		for _, svc := range services {
 			names = append(names, svc.Name)
 		}
+		sort.Strings(names)
 		return names
 	}
 
@@ -221,6 +226,80 @@ func addRestartServicesTasks(st *state.State, queueTask func(task *state.Task),
 	}
 }
 
+func (m *ServiceManager) doQuotaAddSnap(t *state.Task, _ *tomb.Tomb) error {
+	st := t.State()
+	st.Lock()
+	defer st.Unlock()
+
+	perfTimings := state.TimingsForTask(t)
+	defer perfTimings.Save(st)
+
+	snapsup, err := snapstate.TaskSnapSetup(t)
+	if err != nil {
+		return err
+	}
+
+	var quotaName string
+	if err := t.Get("quota-name", &quotaName); err != nil {
+		return fmt.Errorf("internal error: cannot get quota-name: %v", err)
+	}
+
+	allGrps, err := AllQuotas(st)
+	if err != nil {
+		return err
+	}
+
+	qc := QuotaControlAction{
+		Action:    "update",
+		QuotaName: quotaName,
+		AddSnaps:  []string{snapsup.InstanceName()},
+	}
+	grp, allGrps, _, err := quotaUpdate(st, qc, allGrps)
+	if err != nil {
+		return err
+	}
+
+	// ensure service and slices on disk and their states are updated
+	opts := &ensureSnapServicesForGroupOptions{
+		allGrps: allGrps,
+	}
+
+	servicesAffected, err := ensureSnapServicesForGroup(st, t, grp, opts)
+	if err != nil {
+		return err
+	}
+
+	// ensure that if any services are affected, they get their profiles
+	// refreshed immediately as a part of the install change.
+	if len(servicesAffected) > 0 && grp.JournalLimit != nil {
+		ts := state.NewTaskSet()
+		addRefreshProfileTasks(st, func(task *state.Task) {
+			ts.AddTask(task)
+		}, servicesAffected)
+		snapstate.InjectTasks(t, ts)
+	}
+	return nil
+}
+
+func (m *ServiceManager) undoQuotaAddSnap(t *state.Task, _ *tomb.Tomb) error {
+	st := t.State()
+	st.Lock()
+	defer st.Unlock()
+
+	perfTimings := state.TimingsForTask(t)
+	defer perfTimings.Save(st)
+
+	snapsup, err := snapstate.TaskSnapSetup(t)
+	if err != nil {
+		return err
+	}
+
+	if err := EnsureSnapAbsentFromQuota(st, snapsup.InstanceName()); err != nil {
+		return err
+	}
+	return nil
+}
+
 var osutilBootID = osutil.BootID
 
 type quotaStateUpdated struct {
@@ -314,12 +393,23 @@ func quotaCreate(st *state.State, action QuotaControlAction, allGrps map[string]
 		return nil, nil, false, fmt.Errorf("cannot create quota group %q: %v", action.QuotaName, err)
 	}
 
+	// verify we are not trying to add a mixture of services and snaps
+	if len(action.AddSnaps) > 0 && len(action.AddServices) > 0 {
+		return nil, nil, false, fmt.Errorf("cannot mix services and snaps in the same quota group")
+	}
+
 	// make sure the specified snaps exist and aren't currently in another group
-	if err := validateSnapForAddingToGroup(st, action.AddSnaps, action.QuotaName, allGrps); err != nil {
+	if err := validateSnapForAddingToGroup(st, action.AddSnaps, action.QuotaName, parentGrp, allGrps); err != nil {
 		return nil, nil, false, err
 	}
 
-	grp, allGrps, err := internal.CreateQuotaInState(st, action.QuotaName, parentGrp, action.AddSnaps, action.ResourceLimits, allGrps)
+	// if services are provided, the make sure they refer to correct snaps and valid
+	// services.
+	if err := validateSnapServicesForAddingToGroup(st, action.AddServices, action.QuotaName, parentGrp, allGrps); err != nil {
+		return nil, nil, false, err
+	}
+
+	grp, allGrps, err := internal.CreateQuotaInState(st, action.QuotaName, parentGrp, action.AddSnaps, action.AddServices, action.ResourceLimits, allGrps)
 	if err != nil {
 		return nil, nil, false, err
 	}
@@ -344,8 +434,8 @@ func quotaRemove(st *state.State, action QuotaControlAction, allGrps map[string]
 		return nil, nil, false, fmt.Errorf("internal error, AddSnaps option cannot be used with remove action")
 	}
 
-	if action.ResourceLimits.Memory != nil {
-		return nil, nil, false, fmt.Errorf("internal error, MemoryLimit option cannot be used with remove action")
+	if !action.ResourceLimits.Unset() {
+		return nil, nil, false, fmt.Errorf("internal error, quota limit options cannot be used with remove action")
 	}
 
 	// XXX: remove this limitation eventually
@@ -399,8 +489,24 @@ func quotaRemove(st *state.State, action QuotaControlAction, allGrps map[string]
 	return grp, allGrps, refreshProfiles, nil
 }
 
+func validateQuotaLimitsChange(grp *quota.Group, oldLimits, newLimits quota.Resources) error {
+	// Do not allow setting a journal limit on any group which has
+	// services in them. Due to mounts being generated per-snap we cannot
+	// support per-service journal namespaces currently.
+	if newLimits.Journal != nil && len(grp.Services) > 0 {
+		return fmt.Errorf("journal quotas are not supported for individual services")
+	}
+	if err := oldLimits.ValidateChange(newLimits); err != nil {
+		return err
+	}
+	return nil
+}
+
 func quotaUpdateGroupLimits(grp *quota.Group, limits quota.Resources) error {
 	currentQuotas := grp.GetQuotaResources()
+	if err := validateQuotaLimitsChange(grp, currentQuotas, limits); err != nil {
+		return fmt.Errorf("cannot update limits for group %q: %v", grp.Name, err)
+	}
 	if err := currentQuotas.Change(limits); err != nil {
 		return fmt.Errorf("cannot update limits for group %q: %v", grp.Name, err)
 	}
@@ -420,22 +526,27 @@ func quotaUpdate(st *state.State, action QuotaControlAction, allGrps map[string]
 		return nil, nil, false, fmt.Errorf("group %q cannot be moved to a different parent (re-parenting not yet supported)", action.QuotaName)
 	}
 
-	modifiedGrps := []*quota.Group{grp}
-
-	// ensure that the group we are modifying does not contain a mix of snaps and sub-groups
-	// as we no longer support this, and existing quota groups might have this
-	if err := ensureGroupIsNotMixed(action.QuotaName, allGrps); err != nil {
+	// verify we are not trying to add a mixture of services and snaps
+	if err := groupEnsureOnlySnapsOrServices(action.AddSnaps, action.AddServices, grp); err != nil {
 		return nil, nil, false, err
 	}
 
 	// now ensure that all of the snaps mentioned in AddSnaps exist as snaps and
 	// that they aren't already in an existing quota group
-	if err := validateSnapForAddingToGroup(st, action.AddSnaps, action.QuotaName, allGrps); err != nil {
+	parentGrp := allGrps[grp.ParentGroup]
+	if err := validateSnapForAddingToGroup(st, action.AddSnaps, action.QuotaName, parentGrp, allGrps); err != nil {
 		return nil, nil, false, err
 	}
 
-	// append the snaps list in the group
+	// if services are provided, the make sure they refer to correct snaps and valid
+	// services.
+	if err := validateSnapServicesForAddingToGroup(st, action.AddServices, action.QuotaName, parentGrp, allGrps); err != nil {
+		return nil, nil, false, err
+	}
+
+	// append snap list and service list in the group
 	grp.Snaps = append(grp.Snaps, action.AddSnaps...)
+	grp.Services = append(grp.Services, action.AddServices...)
 
 	// store the current status of journal quota, if it changes we need
 	// to refresh the profiles for the snaps in the groups
@@ -447,7 +558,7 @@ func quotaUpdate(st *state.State, action QuotaControlAction, allGrps map[string]
 	}
 
 	// update the quota group state
-	allGrps, err := internal.PatchQuotas(st, modifiedGrps...)
+	allGrps, err := internal.PatchQuotas(st, grp)
 	if err != nil {
 		return nil, nil, false, err
 	}
@@ -466,6 +577,61 @@ type ensureSnapServicesForGroupOptions struct {
 	extraSnaps []string
 }
 
+func snapServiceNames(info *snap.Info) []string {
+	var appNames []string
+	for _, app := range info.Services() {
+		appNames = append(appNames, fmt.Sprintf("%s.%s", info.InstanceName(), app.Name))
+	}
+	return appNames
+}
+
+// affectedSnapServices returns a map of snaps and the services affected
+// by performing changes to a quota group. For groups that contain just snaps, all
+// services are added to the map, for groups that contain specific services, only those
+// services listed in the group are affected, as snaps cannot be in same group as services.
+func affectedSnapServices(st *state.State, grp *quota.Group, opts *ensureSnapServicesForGroupOptions) (map[*snap.Info]*wrappers.SnapServiceOptions, []string, error) {
+	snapSvcMap := map[*snap.Info]*wrappers.SnapServiceOptions{}
+	addSnapToMap := func(sn string) (*snap.Info, error) {
+		info, err := snapstate.CurrentInfo(st, sn)
+		if err != nil {
+			return nil, err
+		}
+		if snapSvcMap[info] != nil {
+			return info, nil
+		}
+		opts, err := SnapServiceOptions(st, info, opts.allGrps)
+		if err != nil {
+			return nil, err
+		}
+		snapSvcMap[info] = opts
+		return info, nil
+	}
+
+	// handle extra snaps also passed here, so that they get included
+	// in any case
+	var affectedServices []string
+	for _, sn := range append(grp.Snaps, opts.extraSnaps...) {
+		info, err := addSnapToMap(sn)
+		if err != nil {
+			return nil, nil, err
+		}
+		affectedServices = append(affectedServices, snapServiceNames(info)...)
+	}
+
+	// if the group is a service group, then grp.Snaps is empty and we
+	// need to get the affected snaps from the service names, which are
+	// of format 'snap.service'
+	for _, svc := range grp.Services {
+		parts := strings.SplitN(svc, ".", 2)
+		snapName := parts[0]
+		if _, err := addSnapToMap(snapName); err != nil {
+			return nil, nil, err
+		}
+		affectedServices = append(affectedServices, svc)
+	}
+	return snapSvcMap, affectedServices, nil
+}
+
 // ensureSnapServicesForGroup will handle updating changes to a given
 // quota group on disk, including re-generating systemd slice files,
 // as well as starting newly created quota groups and stopping and
@@ -482,8 +648,6 @@ func ensureSnapServicesForGroup(st *state.State, t *state.Task, grp *quota.Group
 		return nil, fmt.Errorf("internal error: unset group information for ensuring")
 	}
 
-	allGrps := opts.allGrps
-
 	var meterLocked progress.Meter
 	if t == nil {
 		meterLocked = progress.Null
@@ -492,25 +656,16 @@ func ensureSnapServicesForGroup(st *state.State, t *state.Task, grp *quota.Group
 	}
 
 	// build the map of snap infos to options to provide to EnsureSnapServices
-	snapSvcMap := map[*snap.Info]*wrappers.SnapServiceOptions{}
-	for _, sn := range append(grp.Snaps, opts.extraSnaps...) {
-		info, err := snapstate.CurrentInfo(st, sn)
-		if err != nil {
-			return nil, err
-		}
-
-		opts, err := SnapServiceOptions(st, sn, allGrps)
-		if err != nil {
-			return nil, err
-		}
-
-		snapSvcMap[info] = opts
+	snapSvcMap, affectedServices, err := affectedSnapServices(st, grp, opts)
+	if err != nil {
+		return nil, err
 	}
 
 	// TODO: the following lines should maybe be EnsureOptionsForDevice() or
 	// something since it is duplicated a few places
 	ensureOpts := &wrappers.EnsureSnapServicesOptions{
-		Preseeding: snapdenv.Preseeding(),
+		Preseeding:      snapdenv.Preseeding(),
+		IncludeServices: affectedServices,
 	}
 
 	// set RequireMountedSnapdSnap if we are on UC18+ only
@@ -574,10 +729,16 @@ func ensureSnapServicesForGroup(st *state.State, t *state.Task, grp *quota.Group
 			}
 
 		case "service":
-			// in this case, the only way that a service could have been changed
-			// was if it was moved into or out of a slice, in both cases we need
-			// to restart the service
-			markAppForRestart(app.Snap, app)
+			// When an app has its service file changed, we should restart the app
+			// to take into account that the app might have been moved in/out of a
+			// slice, which means limits may have changed.
+			if app != nil {
+				markAppForRestart(app.Snap, app)
+			}
+
+			// If a quota group has it's service file changed, then it's due to the
+			// journal quota being set. We do not need to do any further changes here
+			// as restart of apps in the journal quota is being handled by case 'journald'
 
 			// TODO: what about sockets and timers? activation units just start
 			// the full unit, so as long as the full unit is restarted we should
@@ -628,6 +789,7 @@ func ensureSnapServicesForGroup(st *state.State, t *state.Task, grp *quota.Group
 	// we need to handle the case where a quota was removed, this will only
 	// happen one at a time and can be identified by the grp provided to us
 	// not existing in the state
+	allGrps := opts.allGrps
 	if _, ok := allGrps[grp.Name]; !ok {
 		// stop the quota group, then remove it
 		if !ensureOpts.Preseeding {
@@ -703,31 +865,37 @@ func ensureSnapServicesStateForGroup(st *state.State, grp *quota.Group, opts *en
 	return restartSnapServices(st, appsToRestartBySnap)
 }
 
-func ensureGroupIsNotMixed(group string, allGrps map[string]*quota.Group) error {
-	grp, ok := allGrps[group]
-	if ok && len(grp.SubGroups) != 0 && len(grp.Snaps) != 0 {
-		return fmt.Errorf("quota group %q has mixed snaps and sub-groups, which is no longer supported; removal and re-creation is necessary to modify it", group)
+// groupEnsureOnlySnapsOrServices is a helper which returns an error if the snaps and/or
+// services we are trying to add would result in a mixed group. A group is considered
+// mixed if:
+// 1. The group would end up containing both snaps and services
+// 2. The group would end up containing both services and groups
+func groupEnsureOnlySnapsOrServices(snapsToAdd, servicesToAdd []string, grp *quota.Group) error {
+	if len(snapsToAdd) > 0 && len(servicesToAdd) > 0 {
+		return fmt.Errorf("cannot mix services and snaps in the same quota group")
+	}
+
+	if (len(grp.Services) > 0 && len(snapsToAdd) > 0) || (len(grp.Snaps) > 0 && len(servicesToAdd) > 0) {
+		return fmt.Errorf("cannot mix services and snaps in the same quota group")
 	}
 	return nil
 }
 
-// ensureGroupHasNoSubgroups returns true if the group has no sub-groups or it doesn't exist,
-// otherwise turns false.
-func ensureGroupHasNoSubgroups(group string, allGrps map[string]*quota.Group) bool {
-	grp, ok := allGrps[group]
-	if ok && len(grp.SubGroups) != 0 {
+// groupIsMixed returns whether a group contains both sub-groups and snaps (i.e is mixed)
+func groupIsMixed(grp *quota.Group) bool {
+	if grp == nil {
 		return false
 	}
-	return true
+	return len(grp.SubGroups) > 0 && len(grp.Snaps) > 0
 }
 
-func validateSnapForAddingToGroup(st *state.State, snaps []string, group string, allGrps map[string]*quota.Group) error {
+func validateSnapForAddingToGroup(st *state.State, snaps []string, group string, parentGroup *quota.Group, allGrps map[string]*quota.Group) error {
 	// With the new quotas we don't support groups that have a mixture of snaps and
 	// subgroups, as this will cause issues with nesting. Groups/subgroups may now
 	// only consist of either snaps or subgroups.
 	if len(snaps) > 0 {
-		if !ensureGroupHasNoSubgroups(group, allGrps) {
-			return fmt.Errorf("cannot mix snaps and sub groups in the group %q", group)
+		if groupIsMixed(parentGroup) {
+			return fmt.Errorf("cannot add snaps to group %q: only services are allowed in this sub-group", group)
 		}
 	}
 
@@ -745,11 +913,102 @@ func validateSnapForAddingToGroup(st *state.State, snaps []string, group string,
 			}
 		}
 	}
-
 	return nil
 }
 
-func quotaControlAffectedSnaps(t *state.Task) (snaps []string, err error) {
+// splitSnapServiceName splits and verifies the snap service reference
+// taken in by the frontend. It expects the format snap.service
+func splitSnapServiceName(name string) (string, string, error) {
+	parts := strings.SplitN(name, ".", 2)
+	if len(parts) != 2 {
+		return "", "", fmt.Errorf("invalid snap service: %s", name)
+	}
+	if err := naming.ValidateSnap(parts[0]); err != nil {
+		return "", "", err
+	}
+	return parts[0], parts[1], naming.ValidateApp(parts[1])
+}
+
+// ensureAppReferenceIsService returns whether the service referred to in the
+// snap is actually a service.
+func ensureAppReferenceIsService(st *state.State, snap, service string) error {
+	snapInfo, err := snapstate.CurrentInfo(st, snap)
+	if err != nil {
+		return err
+	}
+
+	appInfo, ok := snapInfo.Apps[service]
+	if !ok {
+		return fmt.Errorf("invalid service %q", service)
+	}
+	if !appInfo.IsService() {
+		return fmt.Errorf("%q is not a service", service)
+	}
+	return nil
+}
+
+// validateSnapServicesForAddingToGroup verifies that the given services can be added
+// to the provided group. Services come in the format of snap.service, and we do
+// the split and parsing in this function. We need to make sure that the snaps referred
+// are related to the current group, and we make sure that the services are valid service names.
+//
+// We do not allow mixing services and sub-groups. A quota group of service must be
+// the final leaf node in the quota group tree.
+// State must locked prior to calling this function
+func validateSnapServicesForAddingToGroup(st *state.State, services []string, group string, parentGroup *quota.Group, allGrps map[string]*quota.Group) error {
+	if len(services) > 0 {
+		grp, ok := allGrps[group]
+		if ok && len(grp.SubGroups) != 0 {
+			return fmt.Errorf("cannot mix services and sub groups in the group %q", group)
+		}
+
+		// We do not support services in a group with a journal limit. Due to how we generate mounts,
+		// which is currently per-snap, we cannot support individual journal namespaces for services.
+		// So services automatically inherit any journal namespace their parent (the group where the
+		// actual snap is) has set.
+		if ok && grp.JournalLimit != nil {
+			return fmt.Errorf("cannot put services into group %q: journal quotas are not supported for individual services", group)
+		}
+	}
+
+	var svcQuotaMap map[string]*quota.Group
+	if parentGroup != nil {
+		svcQuotaMap = parentGroup.ServiceMap()
+	}
+
+	for _, name := range services {
+		snap, service, err := splitSnapServiceName(name)
+		if err != nil {
+			return err
+		}
+		if err = ensureAppReferenceIsService(st, snap, service); err != nil {
+			return fmt.Errorf("cannot add snap service %q: %v", group, err)
+		}
+		if parentGroup == nil || !strutil.ListContains(parentGroup.Snaps, snap) {
+			return fmt.Errorf("cannot add snap service %q: the snap %q must be in a direct parent group of group %q", service, snap, group)
+		}
+		if serviceGrp := svcQuotaMap[name]; serviceGrp != nil {
+			return fmt.Errorf("cannot add snap service %q: the service is already in group %q", service, serviceGrp.Name)
+		}
+	}
+	return nil
+}
+
+func affectedQuotasForQuotaControl(t *state.Task) (quotas []string, err error) {
+	qcs := []QuotaControlAction{}
+	if err := t.Get("quota-control-actions", &qcs); err != nil {
+		return nil, fmt.Errorf("internal error: cannot get quota-control-actions: %v", err)
+	}
+	quotas = make([]string, 0, len(qcs))
+	for _, qc := range qcs {
+		// TODO: the affected quotas will expand beyond this
+		// if we support reparenting or orphaning
+		quotas = append(quotas, qc.QuotaName)
+	}
+	return quotas, nil
+}
+
+func affectedSnapsForQuotaControl(t *state.Task) (snaps []string, err error) {
 	qcs := []QuotaControlAction{}
 	if err := t.Get("quota-control-actions", &qcs); err != nil {
 		return nil, fmt.Errorf("internal error: cannot get quota-control-actions: %v", err)
@@ -793,3 +1052,11 @@ func quotaControlAffectedSnaps(t *state.Task) (snaps []string, err error) {
 	}
 	return snaps, nil
 }
+
+func affectedQuotasForQuotaAddSnap(t *state.Task) (quotas []string, err error) {
+	var quotaName string
+	if err := t.Get("quota-name", &quotaName); err != nil {
+		return nil, fmt.Errorf("internal error: cannot get quota-name: %v", err)
+	}
+	return []string{quotaName}, nil
+}
