@@ -1,4 +1,5 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
+//go:build !nosecboot
 // +build !nosecboot
 
 /*
@@ -31,9 +32,11 @@ import (
 	"io/ioutil"
 	"os"
 	"path/filepath"
+	"reflect"
 
 	"github.com/canonical/go-tpm2"
 	"github.com/canonical/go-tpm2/linux"
+	"github.com/canonical/go-tpm2/mu"
 	sb "github.com/snapcore/secboot"
 	sb_efi "github.com/snapcore/secboot/efi"
 	sb_tpm2 "github.com/snapcore/secboot/tpm2"
@@ -44,10 +47,12 @@ import (
 	"github.com/snapcore/snapd/bootloader"
 	"github.com/snapcore/snapd/bootloader/efi"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/gadget/device"
 	"github.com/snapcore/snapd/kernel/fde"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/osutil/disks"
 	"github.com/snapcore/snapd/secboot"
+	"github.com/snapcore/snapd/secboot/keys"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/snap/snapfile"
 	"github.com/snapcore/snapd/snap/squashfs"
@@ -77,33 +82,47 @@ func (s *secbootSuite) TestCheckTPMKeySealingSupported(c *C) {
 	efiNotSupported := []uint8(nil)
 	tpmErr := errors.New("TPM error")
 
+	tpmModesBoth := []secboot.TPMProvisionMode{secboot.TPMProvisionFull, secboot.TPMPartialReprovision}
 	type testCase struct {
 		tpmErr     error
 		tpmEnabled bool
+		tpmLockout bool
+		tpmModes   []secboot.TPMProvisionMode
 		sbData     []uint8
 		err        string
 	}
 	for i, tc := range []testCase{
 		// happy case
-		{tpmErr: nil, tpmEnabled: true, sbData: sbEnabled, err: ""},
+		{tpmErr: nil, tpmEnabled: true, tpmLockout: false, tpmModes: tpmModesBoth, sbData: sbEnabled, err: ""},
 		// secure boot EFI var is empty
-		{tpmErr: nil, tpmEnabled: true, sbData: sbEmpty, err: "secure boot variable does not exist"},
+		{tpmErr: nil, tpmEnabled: true, tpmLockout: false, tpmModes: tpmModesBoth, sbData: sbEmpty, err: "secure boot variable does not exist"},
 		// secure boot is disabled
-		{tpmErr: nil, tpmEnabled: true, sbData: sbDisabled, err: "secure boot is disabled"},
+		{tpmErr: nil, tpmEnabled: true, tpmLockout: false, tpmModes: tpmModesBoth, sbData: sbDisabled, err: "secure boot is disabled"},
 		// EFI not supported
-		{tpmErr: nil, tpmEnabled: true, sbData: efiNotSupported, err: "not a supported EFI system"},
+		{tpmErr: nil, tpmEnabled: true, tpmLockout: false, tpmModes: tpmModesBoth, sbData: efiNotSupported, err: "not a supported EFI system"},
 		// TPM connection error
-		{tpmErr: tpmErr, sbData: sbEnabled, err: "cannot connect to TPM device: TPM error"},
+		{tpmErr: tpmErr, sbData: sbEnabled, tpmLockout: false, tpmModes: tpmModesBoth, err: "cannot connect to TPM device: TPM error"},
 		// TPM was detected but it's not enabled
-		{tpmErr: nil, tpmEnabled: false, sbData: sbEnabled, err: "TPM device is not enabled"},
+		{tpmErr: nil, tpmEnabled: false, tpmLockout: false, tpmModes: tpmModesBoth, sbData: sbEnabled, err: "TPM device is not enabled"},
 		// No TPM device
-		{tpmErr: sb_tpm2.ErrNoTPM2Device, sbData: sbEnabled, err: "cannot connect to TPM device: no TPM2 device is available"},
+		{tpmErr: sb_tpm2.ErrNoTPM2Device, tpmLockout: false, tpmModes: tpmModesBoth, sbData: sbEnabled, err: "cannot connect to TPM device: no TPM2 device is available"},
+
+		// In DA lockout mode full provision errors
+		{tpmErr: nil, tpmEnabled: true, tpmLockout: true, tpmModes: []secboot.TPMProvisionMode{secboot.TPMProvisionFull}, sbData: sbEnabled, err: "the TPM is in DA lockout mode"},
+
+		// In DA lockout mode partial provision is fine
+		{tpmErr: nil, tpmEnabled: true, tpmLockout: true, tpmModes: []secboot.TPMProvisionMode{secboot.TPMPartialReprovision}, sbData: sbEnabled, err: ""},
 	} {
-		c.Logf("%d: %v %v %v %q", i, tc.tpmErr, tc.tpmEnabled, tc.sbData, tc.err)
+		c.Logf("%d: %v %v %v %v %q", i, tc.tpmErr, tc.tpmEnabled, tc.tpmModes, tc.sbData, tc.err)
 
 		_, restore := mockSbTPMConnection(c, tc.tpmErr)
 		defer restore()
 
+		restore = secboot.MockSbLockoutAuthSet(func(tpm *sb_tpm2.Connection) bool {
+			return tc.tpmLockout
+		})
+		defer restore()
+
 		restore = secboot.MockIsTPMEnabled(func(tpm *sb_tpm2.Connection) bool {
 			return tc.tpmEnabled
 		})
@@ -116,11 +135,13 @@ func (s *secbootSuite) TestCheckTPMKeySealingSupported(c *C) {
 		restoreEfiVars := efi.MockVars(vars, nil)
 		defer restoreEfiVars()
 
-		err := secboot.CheckTPMKeySealingSupported()
-		if tc.err == "" {
-			c.Assert(err, IsNil)
-		} else {
-			c.Assert(err, ErrorMatches, tc.err)
+		for _, tpmMode := range tc.tpmModes {
+			err := secboot.CheckTPMKeySealingSupported(tpmMode)
+			if tc.err == "" {
+				c.Assert(err, IsNil)
+			} else {
+				c.Assert(err, ErrorMatches, tc.err)
+			}
 		}
 	}
 }
@@ -316,6 +337,86 @@ func (s *secbootSuite) TestLockTPMSealedKeys(c *C) {
 	}
 }
 
+func (s *secbootSuite) TestProvisionForCVM(c *C) {
+	mockTpm, restore := mockSbTPMConnection(c, nil)
+	defer restore()
+
+	restore = secboot.MockIsTPMEnabled(func(tpm *sb_tpm2.Connection) bool {
+		c.Check(tpm, Equals, mockTpm)
+		return true
+	})
+	defer restore()
+
+	expectedTemplate := &tpm2.Public{
+		Type:    tpm2.ObjectTypeRSA,
+		NameAlg: tpm2.HashAlgorithmSHA256,
+		Attrs: tpm2.AttrFixedTPM | tpm2.AttrFixedParent | tpm2.AttrSensitiveDataOrigin | tpm2.AttrUserWithAuth | tpm2.AttrNoDA |
+			tpm2.AttrRestricted | tpm2.AttrDecrypt,
+		Params: &tpm2.PublicParamsU{
+			RSADetail: &tpm2.RSAParams{
+				Symmetric: tpm2.SymDefObject{
+					Algorithm: tpm2.SymObjectAlgorithmAES,
+					KeyBits:   &tpm2.SymKeyBitsU{Sym: 128},
+					Mode:      &tpm2.SymModeU{Sym: tpm2.SymModeCFB}},
+				Scheme:   tpm2.RSAScheme{Scheme: tpm2.RSASchemeNull},
+				KeyBits:  2048,
+				Exponent: 0}}}
+	mu.MustCopyValue(&expectedTemplate, expectedTemplate)
+
+	dir := c.MkDir()
+
+	f, err := os.OpenFile(filepath.Join(dir, "tpm2-srk.tmpl"), os.O_RDWR|os.O_CREATE, 0600)
+	c.Assert(err, IsNil)
+	defer f.Close()
+	mu.MustMarshalToWriter(f, mu.Sized(expectedTemplate))
+
+	provisioningCalls := 0
+	restore = secboot.MockSbTPMEnsureProvisionedWithCustomSRK(func(tpm *sb_tpm2.Connection, mode sb_tpm2.ProvisionMode, lockoutAuth []byte, srkTemplate *tpm2.Public) error {
+		provisioningCalls += 1
+		c.Check(tpm, Equals, mockTpm)
+		c.Check(mode, Equals, sb_tpm2.ProvisionModeWithoutLockout)
+		c.Check(lockoutAuth, IsNil)
+		c.Check(srkTemplate, DeepEquals, expectedTemplate)
+		return nil
+	})
+	defer restore()
+
+	c.Check(secboot.ProvisionForCVM(dir), IsNil)
+	c.Check(provisioningCalls, Equals, 1)
+}
+
+func (s *secbootSuite) TestProvisionForCVMNoTPM(c *C) {
+	_, restore := mockSbTPMConnection(c, sb_tpm2.ErrNoTPM2Device)
+	defer restore()
+
+	restore = secboot.MockSbTPMEnsureProvisionedWithCustomSRK(func(tpm *sb_tpm2.Connection, mode sb_tpm2.ProvisionMode, lockoutAuth []byte, srkTemplate *tpm2.Public) error {
+		c.Error("unexpected provisioning call")
+		return nil
+	})
+	defer restore()
+
+	c.Check(secboot.ProvisionForCVM(c.MkDir()), IsNil)
+}
+
+func (s *secbootSuite) TestProvisionForCVMTPMNotEnabled(c *C) {
+	mockTpm, restore := mockSbTPMConnection(c, nil)
+	defer restore()
+
+	restore = secboot.MockIsTPMEnabled(func(tpm *sb_tpm2.Connection) bool {
+		c.Check(tpm, Equals, mockTpm)
+		return false
+	})
+	defer restore()
+
+	restore = secboot.MockSbTPMEnsureProvisionedWithCustomSRK(func(tpm *sb_tpm2.Connection, mode sb_tpm2.ProvisionMode, lockoutAuth []byte, srkTemplate *tpm2.Public) error {
+		c.Error("unexpected provisioning call")
+		return nil
+	})
+	defer restore()
+
+	c.Check(secboot.ProvisionForCVM(c.MkDir()), IsNil)
+}
+
 func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncrypted(c *C) {
 
 	// setup mock disks to use for locating the partition
@@ -323,30 +424,33 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncrypted(c *C) {
 	// defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
-	mockDiskWithoutAnyDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{},
-	}
+	mockDiskWithoutAnyDev := &disks.MockDiskMapping{}
 
 	mockDiskWithUnencDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"name": "unenc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "name",
+				PartitionUUID:   "unenc-dev-partuuid",
+			},
 		},
 	}
 
 	for idx, tc := range []struct {
 		tpmErr              error
-		keyfile             string // the keyfile to be used to unseal
-		tpmEnabled          bool   // TPM storage and endorsement hierarchies disabled, only relevant if TPM available
-		hasEncdev           bool   // an encrypted device exists
-		rkAllow             bool   // allow recovery key activation
-		rkErr               error  // recovery key unlock error, only relevant if TPM not available
-		activated           bool   // the activation operation succeeded
-		activateErr         error  // the activation error
+		tpmEnabled          bool  // TPM storage and endorsement hierarchies disabled, only relevant if TPM available
+		hasEncdev           bool  // an encrypted device exists
+		rkAllow             bool  // allow recovery key activation
+		rkErr               error // recovery key unlock error, only relevant if TPM not available
+		activated           bool  // the activation operation succeeded
+		activateErr         error // the activation error
 		err                 string
 		skipDiskEnsureCheck bool // whether to check to ensure the mock disk contains the device label
 		expUnlockMethod     secboot.UnlockMethod
@@ -358,14 +462,6 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncrypted(c *C) {
 			activated:       true,
 			disk:            mockDiskWithEncDev,
 			expUnlockMethod: secboot.UnlockedWithSealedKey,
-		}, {
-			// happy case with tpm and encrypted device
-			// with an alternative keyfile
-			tpmEnabled: true, hasEncdev: true,
-			activated:       true,
-			disk:            mockDiskWithEncDev,
-			keyfile:         "some-other-keyfile",
-			expUnlockMethod: secboot.UnlockedWithSealedKey,
 		}, {
 			// device activation fails
 			tpmEnabled: true, hasEncdev: true,
@@ -384,24 +480,9 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncrypted(c *C) {
 			// happy case with tpm and encrypted device, activation
 			// with recovery key
 			tpmEnabled: true, hasEncdev: true, activated: true,
-			activateErr: &sb_tpm2.ActivateWithSealedKeyError{
-				// activation error with nil recovery key error
-				// implies volume activated successfully using
-				// the recovery key,
-				RecoveryKeyUsageErr: nil,
-			},
+			activateErr:     sb.ErrRecoveryKeyUsed,
 			disk:            mockDiskWithEncDev,
 			expUnlockMethod: secboot.UnlockedWithRecoveryKey,
-		}, {
-			// tpm and encrypted device, successful activation, but
-			// recovery key non-nil is an unexpected state
-			tpmEnabled: true, hasEncdev: true, activated: true,
-			activateErr: &sb_tpm2.ActivateWithSealedKeyError{
-				RecoveryKeyUsageErr: fmt.Errorf("unexpected"),
-			},
-			expUnlockMethod: secboot.UnlockStatusUnknown,
-			err:             `internal error: volume activated with unexpected error: .* \(unexpected\)`,
-			disk:            mockDiskWithEncDev,
 		}, {
 			// tpm error, no encrypted device
 			tpmErr: errors.New("tpm error"),
@@ -479,22 +560,33 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncrypted(c *C) {
 		if tc.hasEncdev {
 			fsLabel += "-enc"
 		}
-		partuuid, ok := tc.disk.FilesystemLabelToPartUUID[fsLabel]
+
+		partuuid := ""
 		if !tc.skipDiskEnsureCheck {
-			c.Assert(ok, Equals, true)
+			for _, p := range tc.disk.Structure {
+				if p.FilesystemLabel == fsLabel {
+					partuuid = p.PartitionUUID
+					break
+				}
+			}
+			c.Assert(partuuid, Not(Equals), "", Commentf("didn't find fs label %s in disk", fsLabel))
 		}
+
 		devicePath := filepath.Join("/dev/disk/by-partuuid", partuuid)
 
-		expKeyPath := tc.keyfile
-		if expKeyPath == "" {
-			expKeyPath = "vanilla-keyfile"
-		}
+		keyPath := filepath.Join("test-data", "keyfile")
+		kd, err := sb_tpm2.NewKeyDataFromSealedKeyObjectFile(keyPath)
+		c.Assert(err, IsNil)
+		expectedID, err := kd.UniqueID()
+		c.Assert(err, IsNil)
 
-		restore = secboot.MockSbActivateVolumeWithTPMSealedKey(func(tpm *sb_tpm2.Connection, volumeName, sourceDevicePath,
-			keyPath string, pinReader io.Reader, options *sb.ActivateVolumeOptions) (bool, error) {
+		restore = secboot.MockSbActivateVolumeWithKeyData(func(volumeName, sourceDevicePath string, keyData *sb.KeyData, options *sb.ActivateVolumeOptions) (sb.SnapModelChecker, error) {
 			c.Assert(volumeName, Equals, "name-"+randomUUID)
 			c.Assert(sourceDevicePath, Equals, devicePath)
-			c.Assert(keyPath, Equals, expKeyPath)
+			c.Assert(keyData, NotNil)
+			uID, err := keyData.UniqueID()
+			c.Assert(err, IsNil)
+			c.Check(uID, DeepEquals, expectedID)
 			if tc.rkAllow {
 				c.Assert(*options, DeepEquals, sb.ActivateVolumeOptions{
 					PassphraseTries:  1,
@@ -510,9 +602,9 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncrypted(c *C) {
 				})
 			}
 			if !tc.activated && tc.activateErr == nil {
-				return false, errors.New("activation error")
+				return nil, errors.New("activation error")
 			}
-			return tc.activated, tc.activateErr
+			return nil, tc.activateErr
 		})
 		defer restore()
 
@@ -529,7 +621,7 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncrypted(c *C) {
 		opts := &secboot.UnlockVolumeUsingSealedKeyOptions{
 			AllowRecoveryKey: tc.rkAllow,
 		}
-		unlockRes, err := secboot.UnlockVolumeUsingSealedKeyIfEncrypted(tc.disk, defaultDevice, expKeyPath, opts)
+		unlockRes, err := secboot.UnlockVolumeUsingSealedKeyIfEncrypted(tc.disk, defaultDevice, keyPath, opts)
 		if tc.err == "" {
 			c.Assert(err, IsNil)
 			c.Assert(unlockRes.IsEncrypted, Equals, tc.hasEncdev)
@@ -594,12 +686,12 @@ func (s *secbootSuite) TestEFIImageFromBootFile(c *C) {
 		{
 			// invalid snap file
 			bootFile: bootloader.NewBootFile(existingFile, "rel", bootloader.RoleRecovery),
-			err:      fmt.Sprintf(`"%s/foo" is not a snap or snapdir`, tmpDir),
+			err:      fmt.Sprintf(`cannot process snap or snapdir: cannot read "%s/foo": EOF`, tmpDir),
 		},
 		{
 			// missing snap file
 			bootFile: bootloader.NewBootFile(missingFile, "rel", bootloader.RoleRecovery),
-			err:      fmt.Sprintf(`"%s/bar" is not a snap or snapdir`, tmpDir),
+			err:      fmt.Sprintf(`cannot process snap or snapdir: open %s/bar: no such file or directory`, tmpDir),
 		},
 	} {
 		o, err := secboot.EFIImageFromBootFile(&tc.bootFile)
@@ -612,38 +704,104 @@ func (s *secbootSuite) TestEFIImageFromBootFile(c *C) {
 	}
 }
 
+func (s *secbootSuite) TestProvisionTPM(c *C) {
+	mockErr := errors.New("some error")
+
+	for idx, tc := range []struct {
+		tpmErr            error
+		tpmEnabled        bool
+		mode              secboot.TPMProvisionMode
+		writeLockoutAuth  bool
+		provisioningErr   error
+		provisioningCalls int
+		expectedErr       string
+	}{
+		{
+			tpmErr: mockErr, mode: secboot.TPMProvisionFull,
+			expectedErr: "cannot connect to TPM: some error",
+		}, {
+			tpmEnabled: false, mode: secboot.TPMProvisionFull, expectedErr: "TPM device is not enabled",
+		}, {
+			tpmEnabled: true, mode: secboot.TPMProvisionFull, provisioningErr: mockErr,
+			provisioningCalls: 1, expectedErr: "cannot provision TPM: some error",
+		}, {
+			tpmEnabled: true, mode: secboot.TPMPartialReprovision, provisioningCalls: 0,
+			expectedErr: "cannot read existing lockout auth: open .*/lockout-auth: no such file or directory",
+		},
+		// happy cases
+		{
+			tpmEnabled: true, mode: secboot.TPMProvisionFull, provisioningCalls: 1,
+		}, {
+			tpmEnabled: true, mode: secboot.TPMPartialReprovision, writeLockoutAuth: true,
+			provisioningCalls: 1,
+		},
+	} {
+		c.Logf("tc: %v", idx)
+		d := c.MkDir()
+		tpm, restore := mockSbTPMConnection(c, tc.tpmErr)
+		defer restore()
+
+		// mock TPM enabled check
+		restore = secboot.MockIsTPMEnabled(func(t *sb_tpm2.Connection) bool {
+			return tc.tpmEnabled
+		})
+		defer restore()
+
+		lockoutAuthData := []byte{'l', 'o', 'c', 'k', 'o', 'u', 't', 1, 1, 1, 1, 1, 1, 1, 1, 1}
+		if tc.writeLockoutAuth {
+			c.Assert(ioutil.WriteFile(filepath.Join(d, "lockout-auth"), lockoutAuthData, 0644), IsNil)
+		}
+
+		// mock provisioning
+		provisioningCalls := 0
+		restore = secboot.MockSbTPMEnsureProvisioned(func(t *sb_tpm2.Connection, mode sb_tpm2.ProvisionMode, newLockoutAuth []byte) error {
+			provisioningCalls++
+			c.Assert(t, Equals, tpm)
+			c.Assert(mode, Equals, sb_tpm2.ProvisionModeFull)
+			return tc.provisioningErr
+		})
+		defer restore()
+
+		err := secboot.ProvisionTPM(tc.mode, filepath.Join(d, "lockout-auth"))
+		if tc.expectedErr != "" {
+			c.Assert(err, ErrorMatches, tc.expectedErr)
+		} else {
+			c.Assert(err, IsNil)
+		}
+		c.Check(provisioningCalls, Equals, tc.provisioningCalls)
+	}
+
+}
+
 func (s *secbootSuite) TestSealKey(c *C) {
 	mockErr := errors.New("some error")
 
-	for _, tc := range []struct {
+	for idx, tc := range []struct {
 		tpmErr               error
 		tpmEnabled           bool
 		missingFile          bool
 		badSnapFile          bool
-		skipProvision        bool
 		addEFISbPolicyErr    error
 		addEFIBootManagerErr error
 		addSystemdEFIStubErr error
 		addSnapModelErr      error
 		provisioningErr      error
 		sealErr              error
-		provisioningCalls    int
 		sealCalls            int
 		expectedErr          string
 	}{
 		{tpmErr: mockErr, expectedErr: "cannot connect to TPM: some error"},
 		{tpmEnabled: false, expectedErr: "TPM device is not enabled"},
 		{tpmEnabled: true, missingFile: true, expectedErr: "cannot build EFI image load sequences: file /does/not/exist does not exist"},
-		{tpmEnabled: true, badSnapFile: true, expectedErr: `.*/kernel.snap" is not a snap or snapdir`},
+		{tpmEnabled: true, badSnapFile: true, expectedErr: `cannot build EFI image load sequences: cannot process snap or snapdir: cannot read ".*/kernel.snap": EOF`},
 		{tpmEnabled: true, addEFISbPolicyErr: mockErr, expectedErr: "cannot add EFI secure boot policy profile: some error"},
 		{tpmEnabled: true, addEFIBootManagerErr: mockErr, expectedErr: "cannot add EFI boot manager profile: some error"},
 		{tpmEnabled: true, addSystemdEFIStubErr: mockErr, expectedErr: "cannot add systemd EFI stub profile: some error"},
 		{tpmEnabled: true, addSnapModelErr: mockErr, expectedErr: "cannot add snap model profile: some error"},
-		{tpmEnabled: true, provisioningErr: mockErr, provisioningCalls: 1, expectedErr: "cannot provision TPM: some error"},
-		{tpmEnabled: true, sealErr: mockErr, provisioningCalls: 1, sealCalls: 1, expectedErr: "some error"},
-		{tpmEnabled: true, skipProvision: true, provisioningCalls: 0, sealCalls: 1, expectedErr: ""},
-		{tpmEnabled: true, provisioningCalls: 1, sealCalls: 1, expectedErr: ""},
+		{tpmEnabled: true, sealErr: mockErr, sealCalls: 1, expectedErr: "some error"},
+		{tpmEnabled: true, sealCalls: 1, expectedErr: ""},
 	} {
+		c.Logf("tc: %v", idx)
 		tmpDir := c.MkDir()
 		var mockBF []bootloader.BootFile
 		for _, name := range []string{"a", "b", "c", "d"} {
@@ -701,13 +859,11 @@ func (s *secbootSuite) TestSealKey(c *C) {
 			},
 			TPMPolicyAuthKey:       myAuthKey,
 			TPMPolicyAuthKeyFile:   filepath.Join(tmpDir, "policy-auth-key-file"),
-			TPMLockoutAuthFile:     filepath.Join(tmpDir, "lockout-auth-file"),
-			TPMProvision:           !tc.skipProvision,
 			PCRPolicyCounterHandle: 42,
 		}
 
-		myKey := secboot.EncryptionKey{}
-		myKey2 := secboot.EncryptionKey{}
+		myKey := keys.EncryptionKey{}
+		myKey2 := keys.EncryptionKey{}
 		for i := range myKey {
 			myKey[i] = byte(i)
 			myKey2[i] = byte(128 + i)
@@ -868,17 +1024,6 @@ func (s *secbootSuite) TestSealKey(c *C) {
 		})
 		defer restore()
 
-		// mock provisioning
-		provisioningCalls := 0
-		restore = secboot.MockProvisionTPM(func(t *sb_tpm2.Connection, mode sb_tpm2.ProvisionMode, newLockoutAuth []byte) error {
-			provisioningCalls++
-			c.Assert(t, Equals, tpm)
-			c.Assert(mode, Equals, sb_tpm2.ProvisionModeFull)
-			c.Assert(myParams.TPMLockoutAuthFile, testutil.FilePresent)
-			return tc.provisioningErr
-		})
-		defer restore()
-
 		// mock sealing
 		sealCalls := 0
 		restore = secboot.MockSbSealKeyToTPMMultiple(func(t *sb_tpm2.Connection, kr []*sb_tpm2.SealKeyRequest, params *sb_tpm2.KeyCreationParams) (sb_tpm2.PolicyAuthKey, error) {
@@ -907,9 +1052,7 @@ func (s *secbootSuite) TestSealKey(c *C) {
 		} else {
 			c.Assert(err, ErrorMatches, tc.expectedErr)
 		}
-		c.Assert(provisioningCalls, Equals, tc.provisioningCalls)
 		c.Assert(sealCalls, Equals, tc.sealCalls)
-
 	}
 }
 
@@ -928,10 +1071,12 @@ func (s *secbootSuite) TestResealKey(c *C) {
 		provisioningErr        error
 		resealErr              error
 		resealCalls            int
+		revokeErr              error
+		revokeCalls            int
 		expectedErr            string
 	}{
 		// happy case
-		{tpmEnabled: true, resealCalls: 1, expectedErr: ""},
+		{tpmEnabled: true, resealCalls: 1, revokeCalls: 1, expectedErr: ""},
 
 		// unhappy cases
 		{tpmErr: mockErr, expectedErr: "cannot connect to TPM: some error"},
@@ -943,6 +1088,7 @@ func (s *secbootSuite) TestResealKey(c *C) {
 		{tpmEnabled: true, addSnapModelErr: mockErr, expectedErr: "cannot add snap model profile: some error"},
 		{tpmEnabled: true, readSealedKeyObjectErr: mockErr, expectedErr: "some error"},
 		{tpmEnabled: true, resealErr: mockErr, resealCalls: 1, expectedErr: "some error"},
+		{tpmEnabled: true, revokeErr: errors.New("revoke error"), resealCalls: 1, revokeCalls: 1, expectedErr: "revoke error"},
 	} {
 		mockTPMPolicyAuthKey := []byte{1, 3, 3, 7}
 		mockTPMPolicyAuthKeyFile := filepath.Join(c.MkDir(), "policy-auth-key-file")
@@ -955,6 +1101,9 @@ func (s *secbootSuite) TestResealKey(c *C) {
 			c.Assert(err, IsNil)
 		}
 
+		tmpdir := c.MkDir()
+		keyFile := filepath.Join(tmpdir, "keyfile")
+		keyFile2 := filepath.Join(tmpdir, "keyfile2")
 		myParams := &secboot.ResealKeysParams{
 			ModelParams: []*secboot.SealKeyModelParams{
 				{
@@ -963,7 +1112,7 @@ func (s *secbootSuite) TestResealKey(c *C) {
 					Model:          &asserts.Model{},
 				},
 			},
-			KeyFiles:             []string{"keyfile", "keyfile2"},
+			KeyFiles:             []string{keyFile, keyFile2},
 			TPMPolicyAuthKeyFile: mockTPMPolicyAuthKeyFile,
 		}
 
@@ -976,7 +1125,7 @@ func (s *secbootSuite) TestResealKey(c *C) {
 			// mockSealedKeyObjects, although {},{} would
 			// have been enough as well
 			mockSealedKeyFile := filepath.Join("test-data", "keyfile")
-			mockSealedKeyObject, err := sb_tpm2.ReadSealedKeyObject(mockSealedKeyFile)
+			mockSealedKeyObject, err := sb_tpm2.ReadSealedKeyObjectFromFile(mockSealedKeyFile)
 			c.Assert(err, IsNil)
 			mockSealedKeyObjects = append(mockSealedKeyObjects, mockSealedKeyObject)
 		}
@@ -1047,7 +1196,7 @@ func (s *secbootSuite) TestResealKey(c *C) {
 
 		// mock ReadSealedKeyObject
 		readSealedKeyObjectCalls := 0
-		restore = secboot.MockSbReadSealedKeyObject(func(keyfile string) (*sb_tpm2.SealedKeyObject, error) {
+		restore = secboot.MockSbReadSealedKeyObjectFromFile(func(keyfile string) (*sb_tpm2.SealedKeyObject, error) {
 			readSealedKeyObjectCalls++
 			c.Assert(keyfile, Equals, myParams.KeyFiles[readSealedKeyObjectCalls-1])
 			return mockSealedKeyObjects[readSealedKeyObjectCalls-1], tc.readSealedKeyObjectErr
@@ -1065,6 +1214,16 @@ func (s *secbootSuite) TestResealKey(c *C) {
 			return tc.resealErr
 		})
 		defer restore()
+		// mock PCR protection policy revoke
+		revokeCalls := 0
+		restore = secboot.MockSbSealedKeyObjectRevokeOldPCRProtectionPolicies(func(sko *sb_tpm2.SealedKeyObject, t *sb_tpm2.Connection, authKey sb_tpm2.PolicyAuthKey) error {
+			revokeCalls++
+			c.Assert(sko, Equals, mockSealedKeyObjects[0])
+			c.Assert(t, Equals, tpm)
+			c.Assert(authKey, DeepEquals, sb_tpm2.PolicyAuthKey(mockTPMPolicyAuthKey))
+			return tc.revokeErr
+		})
+		defer restore()
 
 		err = secboot.ResealKeys(myParams)
 		if tc.expectedErr == "" {
@@ -1072,23 +1231,30 @@ func (s *secbootSuite) TestResealKey(c *C) {
 			c.Assert(addEFISbPolicyCalls, Equals, 1)
 			c.Assert(addSystemdEfiStubCalls, Equals, 1)
 			c.Assert(addSnapModelCalls, Equals, 1)
+			c.Assert(keyFile, testutil.FilePresent)
+			c.Assert(keyFile2, testutil.FilePresent)
 		} else {
 			c.Assert(err, ErrorMatches, tc.expectedErr, Commentf("%v", tc))
+			if revokeCalls == 0 {
+				// files were not written out
+				c.Assert(keyFile, testutil.FileAbsent)
+				c.Assert(keyFile2, testutil.FileAbsent)
+			}
 		}
 		c.Assert(resealCalls, Equals, tc.resealCalls)
+		c.Assert(revokeCalls, Equals, tc.revokeCalls)
 	}
 }
 
 func (s *secbootSuite) TestSealKeyNoModelParams(c *C) {
 	myKeys := []secboot.SealKeyRequest{
 		{
-			Key:     secboot.EncryptionKey{},
+			Key:     keys.EncryptionKey{},
 			KeyFile: "keyfile",
 		},
 	}
 	myParams := secboot.SealKeysParams{
 		TPMPolicyAuthKeyFile: "policy-auth-key-file",
-		TPMLockoutAuthFile:   "lockout-auth-file",
 	}
 
 	err := secboot.SealKeys(myKeys, &myParams)
@@ -1126,9 +1292,7 @@ func mockSbTPMConnection(c *C, tpmErr error) (*sb_tpm2.Connection, func()) {
 }
 
 func (s *secbootSuite) TestUnlockEncryptedVolumeUsingKeyBadDisk(c *C) {
-	disk := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{},
-	}
+	disk := &disks.MockDiskMapping{}
 	unlockRes, err := secboot.UnlockEncryptedVolumeUsingKey(disk, "ubuntu-save", []byte("fooo"))
 	c.Assert(err, ErrorMatches, `filesystem label "ubuntu-save-enc" not found`)
 	c.Check(unlockRes, DeepEquals, secboot.UnlockResult{})
@@ -1136,8 +1300,11 @@ func (s *secbootSuite) TestUnlockEncryptedVolumeUsingKeyBadDisk(c *C) {
 
 func (s *secbootSuite) TestUnlockEncryptedVolumeUsingKeyHappy(c *C) {
 	disk := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"ubuntu-save-enc": "123-123-123",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "ubuntu-save-enc",
+				PartitionUUID:   "123-123-123",
+			},
 		},
 	}
 	restore := secboot.MockRandomKernelUUID(func() string {
@@ -1165,8 +1332,11 @@ func (s *secbootSuite) TestUnlockEncryptedVolumeUsingKeyHappy(c *C) {
 
 func (s *secbootSuite) TestUnlockEncryptedVolumeUsingKeyErr(c *C) {
 	disk := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"ubuntu-save-enc": "123-123-123",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "ubuntu-save-enc",
+				PartitionUUID:   "123-123-123",
+			},
 		},
 	}
 	restore := secboot.MockRandomKernelUUID(func() string {
@@ -1200,8 +1370,11 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyErr(
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 	defaultDevice := "name"
@@ -1247,8 +1420,11 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyV1An
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"device-name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "device-name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
@@ -1323,9 +1499,9 @@ func (s *secbootSuite) TestSealKeysWithFDESetupHookHappy(c *C) {
 		return json.Marshal(res)
 	}
 
-	key1 := secboot.EncryptionKey{1, 2, 3, 4}
-	key2 := secboot.EncryptionKey{5, 6, 7, 8}
-	auxKey := secboot.AuxKey{9, 10, 11, 12}
+	key1 := keys.EncryptionKey{1, 2, 3, 4}
+	key2 := keys.EncryptionKey{5, 6, 7, 8}
+	auxKey := keys.AuxKey{9, 10, 11, 12}
 	key1Fn := filepath.Join(tmpdir, "key1.key")
 	key2Fn := filepath.Join(tmpdir, "key2.key")
 	auxKeyFn := filepath.Join(tmpdir, "aux-key")
@@ -1366,8 +1542,8 @@ func (s *secbootSuite) TestSealKeysWithFDESetupHookSad(c *C) {
 		return nil, fmt.Errorf("hook failed")
 	}
 
-	key := secboot.EncryptionKey{1, 2, 3, 4}
-	auxKey := secboot.AuxKey{5, 6, 7, 8}
+	key := keys.EncryptionKey{1, 2, 3, 4}
+	auxKey := keys.AuxKey{5, 6, 7, 8}
 	keyFn := filepath.Join(tmpdir, "key.key")
 	auxKeyFn := filepath.Join(tmpdir, "aux-key")
 	params := secboot.SealKeysWithFDESetupHookParams{
@@ -1384,12 +1560,12 @@ func (s *secbootSuite) TestSealKeysWithFDESetupHookSad(c *C) {
 	c.Check(auxKeyFn, testutil.FileAbsent)
 }
 
-func makeMockDiskKey() secboot.EncryptionKey {
-	return secboot.EncryptionKey{0, 1, 2, 3, 4, 5}
+func makeMockDiskKey() keys.EncryptionKey {
+	return keys.EncryptionKey{0, 1, 2, 3, 4, 5}
 }
 
-func makeMockAuxKey() secboot.AuxKey {
-	return secboot.AuxKey{6, 7, 8, 9}
+func makeMockAuxKey() keys.AuxKey {
+	return keys.AuxKey{6, 7, 8, 9}
 }
 
 func makeMockUnencryptedPayload() []byte {
@@ -1481,8 +1657,11 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyV2(c
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"device-name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "device-name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
@@ -1539,8 +1718,11 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyV2Mo
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"device-name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "device-name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
@@ -1591,8 +1773,11 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyV2Mo
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"device-name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "device-name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
@@ -1641,8 +1826,11 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyV2Al
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"device-name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "device-name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
@@ -1698,8 +1886,11 @@ func (s *secbootSuite) checkV2Key(c *C, keyFn string, prefixToDrop, expectedKey,
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"device-name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "device-name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
@@ -1763,8 +1954,11 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyV1(c
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"device-name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "device-name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
@@ -1817,8 +2011,11 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyBadJ
 	defer restore()
 
 	mockDiskWithEncDev := &disks.MockDiskMapping{
-		FilesystemLabelToPartUUID: map[string]string{
-			"device-name-enc": "enc-dev-partuuid",
+		Structure: []disks.Partition{
+			{
+				FilesystemLabel: "device-name-enc",
+				PartitionUUID:   "enc-dev-partuuid",
+			},
 		},
 	}
 
@@ -1842,3 +2039,140 @@ func (s *secbootSuite) TestUnlockVolumeUsingSealedKeyIfEncryptedFdeRevealKeyBadJ
 
 	c.Check(err, ErrorMatches, `cannot unlock encrypted partition: invalid key data:.*`)
 }
+
+func (s *secbootSuite) TestPCRHandleOfSealedKey(c *C) {
+	d := c.MkDir()
+	h, err := secboot.PCRHandleOfSealedKey(filepath.Join(d, "not-found"))
+	c.Assert(err, ErrorMatches, "cannot open key file: .*/not-found: no such file or directory")
+	c.Assert(h, Equals, uint32(0))
+
+	skf := filepath.Join(d, "sealed-key")
+	// partially valid sealed key with correct header magic
+	c.Assert(ioutil.WriteFile(skf, []byte{0x55, 0x53, 0x4b, 0x24, 1, 1, 1, 'k', 'e', 'y', 1, 1, 1}, 0644), IsNil)
+	h, err = secboot.PCRHandleOfSealedKey(skf)
+	c.Assert(err, ErrorMatches, "(?s)cannot open key file: invalid key data: cannot unmarshal AFIS header: .*")
+	c.Check(h, Equals, uint32(0))
+
+	// TODO simulate the happy case, which needs a real (or at least
+	// partially mocked) sealed key object, which could be obtained using
+	// go-tpm2/testutil, but that has a dependency on an older version of
+	// snapd API and cannot be imported or procure a valid sealed key binary
+	// which unfortunately there are no examples of the secboot/tpm2 test
+	// code
+}
+
+func (s *secbootSuite) TestReleasePCRResourceHandles(c *C) {
+	_, restore := mockSbTPMConnection(c, fmt.Errorf("mock err"))
+	defer restore()
+
+	err := secboot.ReleasePCRResourceHandles(0x1234, 0x2345)
+	c.Assert(err, ErrorMatches, "cannot connect to TPM device: mock err")
+
+	conn, restore := mockSbTPMConnection(c, nil)
+	defer restore()
+
+	var handles []tpm2.Handle
+	restore = secboot.MockTPMReleaseResources(func(tpm *sb_tpm2.Connection, handle tpm2.Handle) error {
+		c.Check(tpm, Equals, conn)
+		handles = append(handles, handle)
+		switch handle {
+		case tpm2.Handle(0xeeeeee):
+			return fmt.Errorf("mock release error 1")
+		case tpm2.Handle(0xeeeeef):
+			return fmt.Errorf("mock release error 2")
+		}
+		return nil
+	})
+	defer restore()
+
+	// many handles
+	err = secboot.ReleasePCRResourceHandles(0x1234, 0x2345)
+	c.Assert(err, IsNil)
+	c.Check(handles, DeepEquals, []tpm2.Handle{
+		tpm2.Handle(0x1234), tpm2.Handle(0x2345),
+	})
+
+	// single handle
+	handles = nil
+	err = secboot.ReleasePCRResourceHandles(0x1234)
+	c.Assert(err, IsNil)
+	c.Check(handles, DeepEquals, []tpm2.Handle{tpm2.Handle(0x1234)})
+
+	// an error case
+	handles = nil
+	err = secboot.ReleasePCRResourceHandles(0x1234, 0xeeeeee, 0x2345, 0xeeeeef)
+	c.Assert(err, ErrorMatches, `
+cannot release TPM resources for 2 handles:
+handle 0xeeeeee: mock release error 1
+handle 0xeeeeef: mock release error 2`[1:])
+	c.Check(handles, DeepEquals, []tpm2.Handle{
+		tpm2.Handle(0x1234), tpm2.Handle(0xeeeeee), tpm2.Handle(0x2345), tpm2.Handle(0xeeeeef),
+	})
+}
+
+func (s *secbootSuite) TestMarkSuccessfulNotEncrypted(c *C) {
+	restore := secboot.MockSbConnectToDefaultTPM(func() (*sb_tpm2.Connection, error) {
+		c.Fatalf("should not get called")
+		return nil, errors.New("boom")
+	})
+	defer restore()
+
+	// device is not encrypted
+	encrypted := device.HasEncryptedMarkerUnder(dirs.SnapFDEDir)
+	c.Assert(encrypted, Equals, false)
+
+	// mark successful returns no error but does not talk to the TPM
+	err := secboot.MarkSuccessful()
+	c.Check(err, IsNil)
+}
+
+func (s *secbootSuite) TestMarkSuccessfulEncryptedTPM(c *C) {
+	s.testMarkSuccessfulEncrypted(c, device.SealingMethodTPM, 1)
+}
+
+func (s *secbootSuite) TestMarkSuccessfulEncryptedFDE(c *C) {
+	s.testMarkSuccessfulEncrypted(c, device.SealingMethodFDESetupHook, 0)
+}
+
+func (s *secbootSuite) testMarkSuccessfulEncrypted(c *C, sealingMethod device.SealingMethod, expectedDaLockResetCalls int) {
+	_, restore := mockSbTPMConnection(c, nil)
+	defer restore()
+
+	// device is encrypted
+	err := os.MkdirAll(dirs.SnapFDEDir, 0700)
+	c.Assert(err, IsNil)
+	saveFDEDir := dirs.SnapFDEDirUnderSave(dirs.SnapSaveDir)
+	err = os.MkdirAll(saveFDEDir, 0700)
+	c.Assert(err, IsNil)
+
+	err = device.StampSealedKeys(dirs.GlobalRootDir, sealingMethod)
+	c.Assert(err, IsNil)
+
+	// write fake lockout auth
+	lockoutAuthValue := []byte("tpm-lockout-auth-key")
+	err = ioutil.WriteFile(filepath.Join(saveFDEDir, "tpm-lockout-auth"), lockoutAuthValue, 0600)
+	c.Assert(err, IsNil)
+
+	daLockResetCalls := 0
+	restore = secboot.MockSbTPMDictionaryAttackLockReset(func(tpm *sb_tpm2.Connection, lockContext tpm2.ResourceContext, lockContextAuthSession tpm2.SessionContext, sessions ...tpm2.SessionContext) error {
+		daLockResetCalls++
+		// Below this code pokes at the private data from
+		//   github.com/canonical/go-tpm2/resources.go
+		//   type resourceContext struct {
+		//     ...
+		//     authValue []byte
+		//   }
+		// there is no exported API to get the auth value. If go-tpm2
+		// starts chaning it's probably not worth updating this
+		// part of the test and it can just get removed.
+		fv := reflect.ValueOf(lockContext).Elem().FieldByName("authValue")
+		c.Check(fv.Bytes(), DeepEquals, lockoutAuthValue)
+		return nil
+	})
+	defer restore()
+
+	err = secboot.MarkSuccessful()
+	c.Check(err, IsNil)
+
+	c.Check(daLockResetCalls, Equals, expectedDaLockResetCalls)
+}
