@@ -21,23 +21,35 @@ package wrappers
 
 import (
 	"time"
+
+	"github.com/snapcore/snapd/osutil"
 )
 
 // some internal helper exposed for testing
 var (
 	// services
 	GenerateSnapServiceFile = generateSnapServiceFile
+	GenerateSnapSocketFiles = generateSnapSocketFiles
 	GenerateSnapTimerFile   = generateSnapTimerFile
 
+	// dbus
+	GenerateDBusActivationFile = generateDBusActivationFile
+
 	// desktop
 	SanitizeDesktopFile    = sanitizeDesktopFile
 	RewriteExecLine        = rewriteExecLine
+	RewriteIconLine        = rewriteIconLine
 	IsValidDesktopFileLine = isValidDesktopFileLine
 
 	// timers
 	GenerateOnCalendarSchedules = generateOnCalendarSchedules
+
+	// icons
+	FindIconFiles = findIconFiles
 )
 
+type GenerateSnapServicesOptions = generateSnapServicesOptions
+
 func MockKillWait(wait time.Duration) (restore func()) {
 	oldKillWait := killWait
 	killWait = wait
@@ -45,3 +57,11 @@ func MockKillWait(wait time.Duration) (restore func()) {
 		killWait = oldKillWait
 	}
 }
+
+func MockEnsureDirState(f func(dir string, glob string, content map[string]osutil.FileState) (changed, removed []string, err error)) (restore func()) {
+	oldEnsureDirState := ensureDirState
+	ensureDirState = f
+	return func() {
+		ensureDirState = oldEnsureDirState
+	}
+}
